import os

import pytest
from fastmcp.tools import Tool as FastMCPTool
from mcp.types import Tool

from fastmcp_agents.conversation.types import Conversation, SystemConversationEntry, UserConversationEntry
from fastmcp_agents.errors.llm_link import ModelDoesNotSupportFunctionCallingError
from fastmcp_agents.llm_link.litellm import AsyncLitellmLLMLink


@pytest.fixture
def mock_mcp_tool():
    return Tool(
        name="test_tool",
        description="A test tool",
        inputSchema={
            "type": "object",
            "properties": {"test_param": {"type": "string", "description": "A test parameter"}},
            "required": ["test_param"],
        },
    )


@pytest.fixture
def mock_fastmcp_tool():
    def test_tool():
        pass

    return FastMCPTool.from_function(fn=test_tool)


@pytest.fixture
def conversation():
    conv = Conversation()
    conv = conv.append(SystemConversationEntry(content="You are a test assistant"))
    return conv.append(UserConversationEntry(content="Test message"))


def test_llm_link_initialization_with_invalid_model():
    """Test that LLM link initialization fails with an invalid model."""
    with pytest.raises(ModelDoesNotSupportFunctionCallingError):
        AsyncLitellmLLMLink(model="invalid-model")


async def test_llm_link_completion_with_tools(mock_fastmcp_tool, conversation):
    """Test that LLM link can make a completion with tools."""
    # Use a real model that supports function calling
    model = os.getenv("MODEL", "gpt-3.5-turbo")
    llm_link = AsyncLitellmLLMLink(model=model)

    assistant_conversation_entry = await llm_link.async_completion(conversation=conversation, fastmcp_tools=[mock_fastmcp_tool])

    tool_calls = assistant_conversation_entry.tool_calls
    # Verify tool calls were generated
    assert isinstance(tool_calls, list)
    if tool_calls:  # Some models might not always generate tool calls
        assert all(hasattr(call, "name") for call in tool_calls)
        assert all(hasattr(call, "arguments") for call in tool_calls)



"""Functions for loading tool transformation configurations."""

import json
from pathlib import Path
from typing import Any

import yaml

from fastmcp_agents.vendored.tool_transformer.models import ToolOverride


def overrides_from_dict(obj: dict[str, Any]) -> dict[str, ToolOverride]:
    return {tool_name: ToolOverride.model_validate(tool_override) for tool_name, tool_override in obj.items()}


def overrides_from_yaml(yaml_str: str) -> dict[str, ToolOverride]:
    return overrides_from_dict(yaml.safe_load(yaml_str))


def overrides_from_yaml_file(yaml_file: Path) -> dict[str, ToolOverride]:
    with Path(yaml_file).open(encoding="utf-8") as f:
        return overrides_from_yaml(yaml_str=f.read())


def overrides_from_json(json_str: str) -> dict[str, ToolOverride]:
    return overrides_from_dict(json.loads(json_str))


def overrides_from_json_file(json_file: Path) -> dict[str, ToolOverride]:
    with Path(json_file).open(encoding="utf-8") as f:
        return overrides_from_json(f.read())



"""Pydantic models for the CLI."""

from typing import TYPE_CHECKING, Any, Literal

from fastmcp import Client, FastMCP
from fastmcp.server.proxy import ProxyTool
from fastmcp.tools import Tool as FastMCPTool
from fastmcp.utilities.mcp_config import MCPConfig, RemoteMCPServer, StdioMCPServer
from pydantic import BaseModel, Field

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.memory.ephemeral import EphemeralMemory
from fastmcp_agents.errors.cli import MCPServerError
from fastmcp_agents.observability.logging import BASE_LOGGER
from fastmcp_agents.vendored.tool_transformer.models import ToolOverride

if TYPE_CHECKING:
    from collections.abc import Callable

logger = BASE_LOGGER.getChild("agent.loader")


class ServerSettings(BaseModel):
    """Settings for the server."""

    transport: Literal["stdio", "sse", "streamable-http"] = Field(..., description="The transport to use for the server.")
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = Field(..., description="The log level to use for the server.")
    agent_only: bool = Field(default=False, description="Whether to only run the agents.")
    tool_only: bool = Field(default=False, description="Whether to only run the tools.")


class BaseExtraTool(BaseModel):
    """A set of overrides for a tool."""

    name: str = Field(..., description="The name of the tool.")
    description: str = Field(..., description="The description of the tool.")

    def to_fastmcp_tool(self) -> FastMCPTool:
        msg = "Subclasses must implement this method."
        raise NotImplementedError(msg)


class StaticExtraTool(BaseExtraTool):
    """A tool that returns a static string."""

    returns: str = Field(..., description="The string to return from the tool.")

    def to_fastmcp_tool(self) -> FastMCPTool:
        """Convert the tool to a FastMCP tool."""

        def tool_fn() -> str:
            return self.returns

        return FastMCPTool.from_function(
            fn=tool_fn,
            name=self.name,
            description=self.description,
        )


ExtraToolTypes = StaticExtraTool


async def _wrap_client_tools(client: Client, tool_overrides: dict[str, ToolOverride]) -> list[FastMCPTool]:
    """Wrap the client's tools with the tool overrides."""

    async with client:
        client_tools = await client.list_tools()

    proxied_client_tools = [await ProxyTool.from_client(client, client_tool) for client_tool in client_tools]

    return [
        tool_overrides[proxied_client_tool.name].apply_to_tool(proxied_client_tool)
        if proxied_client_tool.name in tool_overrides
        else proxied_client_tool
        for proxied_client_tool in proxied_client_tools
    ]


class OverriddenStdioMCPServer(StdioMCPServer):
    """A Stdio server with overridden tools."""

    tools: dict[str, ToolOverride] = Field(default_factory=dict[str, ToolOverride])

    def to_fastmcp_client(self, init_timeout: float = 20.0) -> Client:
        """Convert the server to a FastMCP client."""
        transport = self.to_transport()
        transport.keep_alive = True
        return Client(transport=transport, init_timeout=init_timeout)

    async def to_wrapped_client(self) -> tuple[Client, list[FastMCPTool]]:
        """Convert the server to a wrapped FastMCP client."""
        client = self.to_fastmcp_client()
        return client, await _wrap_client_tools(client, self.tools)


class OverriddenRemoteMCPServer(RemoteMCPServer):
    """A Remote server with overridden tools."""

    tools: dict[str, ToolOverride] = Field(default_factory=dict)

    def to_fastmcp_client(self, init_timeout: float = 10.0) -> Client:
        """Convert the server to a FastMCP client."""
        return Client(transport=self.to_transport(), init_timeout=init_timeout)

    async def to_wrapped_client(self) -> tuple[Client, list[FastMCPTool]]:
        """Convert the server to a wrapped FastMCP client."""
        client = self.to_fastmcp_client()
        return client, await _wrap_client_tools(client, self.tools)


class OverriddenMCPConfig(MCPConfig):
    """A MCP config with overridden tools."""

    mcpServers: dict[str, OverriddenStdioMCPServer | OverriddenRemoteMCPServer] = Field(default_factory=dict)  # type: ignore # noqa: N815


class AgentModel(BaseModel):
    """A model for an agent."""

    type: Literal["fastmcp"] = Field(default="fastmcp", description="The type of agent.")
    name: str = Field(..., description="The name of the agent.")
    description: str = Field(..., description="The description of the agent.")
    default_instructions: str = Field(..., description="The default instructions to provide to the agent.")
    model: str | None = Field(default=None, description="The GenAI model to use for the agent.")
    allowed_tools: list[str] | None = Field(None, description="An optional list of the tools to provide to the agent.")
    blocked_tools: list[str] | None = Field(None, description="An optional list of the tools to block from the agent.")

    @classmethod
    def to_fastmcp_agent(cls, agent_model: "AgentModel", tools: list[FastMCPTool]) -> FastMCPAgent:
        """Convert the agent model to a FastMCP agent."""
        agent_tools = list(tools)

        if agent_model.allowed_tools:
            agent_tools = [tool for tool in agent_tools if tool.name in agent_model.allowed_tools]

        if agent_model.blocked_tools:
            agent_tools = [tool for tool in agent_tools if tool.name not in agent_model.blocked_tools]

        logger.debug(f"Agent {agent_model.name} will have access to the following tools: {agent_tools}")

        return FastMCPAgent(
            name=agent_model.name,
            description=agent_model.description,
            system_prompt=agent_model.default_instructions,
            default_tools=agent_tools,
            memory_factory=EphemeralMemory,
        )

    @classmethod
    def to_fastmcp_tool(cls, agent_model: "AgentModel", tools: list[FastMCPTool]) -> tuple[FastMCPAgent, FastMCPTool]:
        """Convert the agent model to a FastMCP tool."""
        agent = cls.to_fastmcp_agent(agent_model, tools)

        return agent, FastMCPTool.from_function(fn=agent.currate, name=agent.name, description=agent.description)


class AugmentedServerModel(BaseModel):
    """A full configuration for an augmented server with Agents and extra tools."""

    agents: list[AgentModel] = Field(default_factory=list)
    mcpServers: dict[str, OverriddenStdioMCPServer | OverriddenRemoteMCPServer] = Field(default_factory=dict)  # noqa: N815
    tools: list[ExtraToolTypes] = Field(default_factory=list)

    async def to_fastmcp_server(self, server_settings: ServerSettings) -> tuple[list[FastMCPAgent], list[Client], FastMCP]:
        """Convert the server model to a FastMCP server, agents, and tools."""

        wrapped_mcp_clients: list[tuple[Client, list[FastMCPTool]]] = []

        for name, server in self.mcpServers.items():
            try:
                wrapped_mcp_clients.append(await server.to_wrapped_client())
            except Exception as e:  # noqa: PERF203
                raise MCPServerError(name, e) from e

        mcp_tools = [tool for _, tools in wrapped_mcp_clients for tool in tools]
        extra_tools = [extra_tool.to_fastmcp_tool() for extra_tool in self.tools]
        all_standard_tools: list[FastMCPTool] = mcp_tools + extra_tools

        exposed_tools: list[FastMCPTool | Callable[..., Any]] = []
        fastmcp_agents: list[FastMCPAgent] = []

        if not server_settings.tool_only:
            for agent_model in self.agents:
                fastmcp_agent, agent_tool = AgentModel.to_fastmcp_tool(agent_model, all_standard_tools)

                fastmcp_agents.append(fastmcp_agent)
                exposed_tools.append(agent_tool)

        if not server_settings.agent_only:
            exposed_tools.extend(all_standard_tools)

        return fastmcp_agents, [client for client, _ in wrapped_mcp_clients], FastMCP(name="augmented-server", tools=exposed_tools)



"""Base custom errors for FastMCP Agents."""


class FastMCPAgentsError(Exception):
    """Base class for all FastMCPAgents errors."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ContributionsWelcomeError(FastMCPAgentsError):
    """Raised when someone else should work on a feature."""

    def __init__(self, feature: str):
        self.message = f"We would welcome a contribution implementing {feature}"
        super().__init__(self.message)


class UnknownToolCallError(FastMCPAgentsError):
    """Raised when a tool call is not known."""

    def __init__(self, tool_name: str, extra_info: str | None = None):
        self.message = f"Unknown tool call: {tool_name}"
        if extra_info:
            self.message += f" {extra_info}"
        super().__init__(self.message)


class UnsupportedFeatureError(FastMCPAgentsError):
    """Raised when a feature is not supported."""

    def __init__(self, feature: str):
        self.message = f"FastMCPAgents does not support the feature: {feature}"
        super().__init__(self.message)


class NoResponseError(FastMCPAgentsError):
    """Raised when a response is not received."""

    def __init__(self, missing_item: str, model: str):
        self.message = f"{missing_item} was missing in the response from {model}"
        super().__init__(self.message)



"""Pydantic models for tool transformation."""

from collections.abc import Awaitable, Callable
from copy import deepcopy
from typing import Annotated, Any, Generic, Literal, Protocol, TypeVar, runtime_checkable

from fastmcp.exceptions import ToolError
from fastmcp.tools import Tool as FastMCPTool
from jsonschema import ValidationError, validate
from mcp.types import EmbeddedResource, ImageContent, TextContent
from pydantic import BaseModel, ConfigDict, Field, model_validator

from fastmcp_agents.vendored.tool_transformer.errors import (
    ToolParameterAlreadyExistsError,
    ToolParameterNotFoundError,
    ToolParameterOverrideError,
)

T = TypeVar("T", bound=bool | int | float | str)  # | dict[str, Any] | BaseModel)


class ToolParameter(BaseModel, Generic[T]):
    """A base class for a tool parameter."""

    name: str = Field(description="The name of the parameter.")
    description: str | None = Field(default=None, description="The description of the parameter.")
    append_description: str | None = Field(default=None, description="A description to append to the parameter's description.")
    required: bool = Field(default=False, description="Whether the parameter is required.")
    constant: T | None = Field(default=None, description="A constant value for the parameter.")
    default: T | None = Field(default=None, description="A default value for the parameter.")

    @model_validator(mode="after")
    def validate_default_and_constant(self):
        if self.default is not None and self.constant is not None:
            msg = "Cannot set both default and constant for a parameter."
            raise ToolParameterOverrideError(msg)

        return self

    @classmethod
    def _get_jsonschema_type(cls) -> str:
        msg = "Subclasses must implement this method."
        raise NotImplementedError(msg)

    def add_into_schema(self, schema: dict[str, Any]) -> dict[str, Any]:
        """
        Adds the parameter into the schema.

        Args:
            schema: The schema of the tool.

        Returns:
            The updated schema. Does not modify the schema in place.
        """

        working_schema = deepcopy(schema)

        if not working_schema.get("properties"):
            working_schema["properties"] = {}

        if self.name in working_schema["properties"]:
            raise ToolParameterAlreadyExistsError(self.name)

        new_parameter: dict[str, Any] = {
            "type": self._get_jsonschema_type(),
            "title": self.name,
        }

        if self.description is not None:
            new_parameter["description"] = self.description

        if self.default is not None:
            new_parameter["default"] = self.default

        if self.constant is not None:
            new_parameter["const"] = self.constant

        working_schema["properties"][self.name] = new_parameter

        if self.required:
            working_schema.setdefault("required", []).append(self.name)

        return working_schema

    def merge_into_schema(self, schema: dict[str, Any]) -> dict[str, Any]:
        """
        Merges the parameter into the schema.

        Args:
            schema: The schema of the tool.

        Returns:
            The updated schema. Also modifies the schema in place.
        """
        working_schema = deepcopy(schema)

        if not working_schema.get("properties"):
            working_schema["properties"] = {}

        if self.name not in working_schema["properties"]:
            raise ToolParameterNotFoundError(self.name)

        parameter_schema = working_schema["properties"][self.name]

        if parameter_schema.get("description") is not None and self.append_description is not None:
            parameter_schema["description"] += "\n" + self.append_description

        if self.description is not None:
            parameter_schema["description"] = self.description

        if self.default is not None:
            parameter_schema["default"] = self.default

        if self.constant is not None:
            parameter_schema["const"] = self.constant

        if self.required and self.name not in working_schema.get("required", []):
            working_schema.setdefault("required", []).append(self.name)

        return working_schema

    def apply_to_schema(
        self,
        schema: dict[str, Any],
        existing_parameter_behavior: Literal["error", "merge"] = "error",
    ) -> dict[str, Any]:
        """
        Applies the parameter to the schema.

        Args:
            schema: The schema of the tool.
            existing_parameter_behavior: The behavior to use when the parameter already exists in the schema.

        """

        already_in_schema = self.name in schema["properties"]

        if not already_in_schema:
            working_schema = self.add_into_schema(schema)
        else:
            if existing_parameter_behavior == "error":
                raise ToolParameterAlreadyExistsError(self.name)
            working_schema = self.merge_into_schema(schema)

        return working_schema


class IntToolParameter(ToolParameter[int]):
    """A parameter that is an integer."""

    @classmethod
    def _get_jsonschema_type(cls) -> str:
        return "number"


class FloatToolParameter(ToolParameter[float]):
    """A parameter that is a float."""

    @classmethod
    def _get_jsonschema_type(cls) -> str:
        return "number"


class StringToolParameter(ToolParameter[str]):
    """A parameter that is a string."""

    @classmethod
    def _get_jsonschema_type(cls) -> str:
        return "string"


class BooleanToolParameter(ToolParameter[bool]):
    """A parameter that is a boolean."""

    @classmethod
    def _get_jsonschema_type(cls) -> str:
        return "boolean"


# class DictToolParameter(ToolParameter[dict[str, Any]]):
#     """A parameter that is an object."""

#     type: dict[str, Any]

#     @classmethod
#     def _get_jsonschema_type(cls) -> str:
#         return "object"


# class BaseModelToolParameter(ToolParameter[BaseModel]):
#     """A parameter that is a model."""

#     type: BaseModel

#     @classmethod
#     def _get_jsonschema_type(cls) -> str:
#         return "object"


ToolParameterTypes = (
    IntToolParameter
    | FloatToolParameter
    | StringToolParameter
    | BooleanToolParameter
    # | DictToolParameter
    # | BaseModelToolParameter
    | ToolParameter
)


@runtime_checkable
class PostToolCallHookProtocol(Protocol):
    async def __call__(
        self,
        response: list[TextContent | ImageContent | EmbeddedResource],
        tool_args: dict[str, Any],
        hook_args: dict[str, Any],
    ) -> None: ...


@runtime_checkable
class PreToolCallHookProtocol(Protocol):
    async def __call__(
        self,
        tool_args: dict[str, Any],
        hook_args: dict[str, Any],
    ) -> None: ...


HookProtocolTypes = PreToolCallHookProtocol | PostToolCallHookProtocol


class ToolOverride(BaseModel):
    """A base class for a tool override that can be used to override a tool."""

    name: str | None = Field(default=None, description="The name of the tool to override.")
    description: str | None = Field(default=None, description="The description of the tool to override.")
    append_description: str | None = Field(default=None, description="A description to append to the tool's description.")
    parameter_overrides: list[ToolParameterTypes] = Field(
        default_factory=list,
        description="A list of parameter overrides to apply to the tool.",
    )
    hook_parameters: list[ToolParameterTypes] = Field(
        default_factory=list,
        description="A list of hook parameters to apply to the tool.",
    )
    pre_call_hook: PreToolCallHookProtocol | None = Field(
        default=None,
        description="A hook to call before the tool is called.",
        exclude=True,
    )
    post_call_hook: PostToolCallHookProtocol | None = Field(
        default=None,
        description="A hook to call after the tool is called.",
        exclude=True,
    )

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def apply_to_tool(
        self,
        tool: FastMCPTool,
    ) -> FastMCPTool:
        return _transform_tool(tool, self)


def _extract_hook_args(
    extra_parameters: list[ToolParameterTypes],
    tool_call_kwargs: dict[str, Any],
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Extracts the hook arguments from the tool call kwargs.
    """
    hook_args: dict[str, Any] = {}
    tool_call_args: dict[str, Any] = deepcopy(tool_call_kwargs)

    extra_parameters_by_name = {extra_parameter.name: extra_parameter for extra_parameter in extra_parameters}

    for tool_arg_name in tool_call_kwargs:
        if tool_arg_name in extra_parameters_by_name:
            hook_args[tool_arg_name] = tool_call_args.pop(tool_arg_name)

    return hook_args, tool_call_args


def _apply_hook_parameters(
    schema: dict[str, Any],
    hook_parameters: list[ToolParameterTypes],
) -> dict[str, Any]:
    """Applies extra parameters for hooks to the tool parameters schema."""
    transformed_schema = deepcopy(schema)

    for hook_parameter in hook_parameters:
        transformed_schema = hook_parameter.add_into_schema(transformed_schema)

    return transformed_schema


def _apply_parameter_overrides(
    schema: dict[str, Any],
    overrides: list[ToolParameterTypes],
) -> dict[str, Any]:
    """
    Applies parameter overrides to a schema.
    """
    transformed_schema = deepcopy(schema)

    for override in overrides:
        transformed_schema = override.merge_into_schema(transformed_schema)

    return transformed_schema


def _create_transformed_function(
    original_tool_run_method: Callable[..., Awaitable[list[TextContent | ImageContent | EmbeddedResource]]],
    fn_transformed_parameters_schema: dict[str, Any],
    fn_hook_parameters_list: list[ToolParameterTypes],
    fn_pre_call_hook: PreToolCallHookProtocol | None,
    fn_post_call_hook: PostToolCallHookProtocol | None,
) -> Callable[..., Awaitable[list[TextContent | ImageContent | EmbeddedResource]]]:
    """Factory function to create the transformed tool's core async callable."""

    async def transformed_fn(
        **kwargs: Any,
    ) -> list[TextContent | ImageContent | EmbeddedResource]:
        try:
            validate(kwargs, fn_transformed_parameters_schema)
        except ValidationError as e:
            msg = f"Provided arguments for {e.json_path} are invalid: {e.message}"
            raise ToolError(msg) from e

        hook_args, tool_call_kwargs = _extract_hook_args(fn_hook_parameters_list, kwargs)

        if fn_pre_call_hook:
            await fn_pre_call_hook(tool_call_kwargs, hook_args)

        response = await original_tool_run_method(arguments=tool_call_kwargs)

        if fn_post_call_hook:
            await fn_post_call_hook(response, tool_call_kwargs, hook_args)

        return response

    return transformed_fn


def _transform_tool(
    tool: FastMCPTool,
    override: Annotated[ToolOverride, "Tool overrides for the tool."],
) -> FastMCPTool:
    transformed_parameters: dict[str, Any] = deepcopy(tool.parameters)

    transformed_parameters = _apply_hook_parameters(schema=transformed_parameters, hook_parameters=override.hook_parameters)
    transformed_parameters = _apply_parameter_overrides(schema=transformed_parameters, overrides=override.parameter_overrides)

    transformed_fn_callable = _create_transformed_function(
        original_tool_run_method=tool.run,
        fn_transformed_parameters_schema=transformed_parameters,
        fn_hook_parameters_list=override.hook_parameters,
        fn_pre_call_hook=override.pre_call_hook,
        fn_post_call_hook=override.post_call_hook,
    )

    transformed_name: str = override.name or tool.name
    transformed_description: str | None = tool.description

    if transformed_description and override.append_description is not None:
        transformed_description += "\n" + override.append_description

    if override.description is not None:
        transformed_description = override.description

    return FastMCPTool(
        fn=transformed_fn_callable,
        name=transformed_name,
        description=transformed_description,
        parameters=transformed_parameters,
        tags=tool.tags,
        annotations=tool.annotations,
        serializer=tool.serializer,
        exclude_args=tool.exclude_args,
    )



import os

import pytest
from mcp.types import Tool

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.memory.base import PrivateMemoryFactory
from fastmcp_agents.conversation.memory.ephemeral import EphemeralMemory
from fastmcp_agents.conversation.types import Conversation, SystemConversationEntry, UserConversationEntry
from fastmcp_agents.llm_link.litellm import AsyncLitellmLLMLink

MODEL = os.getenv("MODEL")

if not MODEL:
    msg = "MODEL environment variable is not set"
    raise ValueError(msg)


@pytest.fixture
def mock_tool():
    return Tool(
        name="test_tool",
        description="A test tool",
        inputSchema={
            "type": "object",
            "properties": {"test_param": {"type": "string", "description": "A test parameter"}},
            "required": ["test_param"],
        },
    )


@pytest.fixture
def llm_link():
    return AsyncLitellmLLMLink(model="gpt-3.5-turbo")


@pytest.fixture
def agent(llm_link, mock_tool):
    return FastMCPAgent(name="test_agent", description="A test agent", default_tools=[mock_tool], llm_link=llm_link)


def test_agent_initialization(agent):
    """Test that an agent can be initialized with the correct properties."""
    assert agent.name == "test_agent"
    assert agent.description == "A test agent"
    assert len(agent.default_tools) == 1
    assert agent.default_tools[0].name == "test_tool"
    assert agent.llm_link is not None


def test_agent_system_prompt_formatting():
    """Test that the system prompt is correctly formatted with agent details."""
    agent = FastMCPAgent(name="test_agent", description="A test agent that can do things", llm_link=AsyncLitellmLLMLink())

    # The system prompt should contain the agent name and description
    memory = PrivateMemoryFactory(memory_class=EphemeralMemory)()
    conversation = agent._prepare_conversation(memory=memory, instructions="Custom Instructions")
    assert isinstance(conversation, Conversation)
    assert len(conversation.entries) == 2
    assert isinstance(conversation.entries[0], SystemConversationEntry)
    assert "test_agent" in conversation.entries[0].content
    assert "A test agent that can do things" in conversation.entries[0].content
    assert isinstance(conversation.entries[1], UserConversationEntry)
    assert "Custom Instructions" in conversation.entries[1].content


def test_agent_with_custom_system_prompt():
    """Test that an agent can be initialized with a custom system prompt."""
    custom_prompt = "You are a specialized test agent"
    agent = FastMCPAgent(name="test_agent", description="A test agent", system_prompt=custom_prompt, llm_link=AsyncLitellmLLMLink())

    memory = PrivateMemoryFactory(memory_class=EphemeralMemory)()
    conversation = agent._prepare_conversation(memory=memory, instructions="Custom Instructions")
    assert isinstance(conversation, Conversation)
    assert len(conversation.entries) == 2
    assert isinstance(conversation.entries[0], SystemConversationEntry)
    assert "You are a specialized test agent" in conversation.entries[0].content
    assert isinstance(conversation.entries[1], UserConversationEntry)
    assert "Custom Instructions" in conversation.entries[1].content


def test_agent_instantiation_step_limit():
    """Test that the agent respects the step limit."""
    agent = FastMCPAgent(name="test_agent", description="A test agent", step_limit=5, llm_link=AsyncLitellmLLMLink())

    assert agent.step_limit == 5


def test_agent_instantiation_max_parallel_tool_calls():
    """Test that the agent respects the max parallel tool calls limit."""
    agent = FastMCPAgent(name="test_agent", description="A test agent", max_parallel_tool_calls=3, llm_link=AsyncLitellmLLMLink())

    assert agent.max_parallel_tool_calls == 3



"""Utility functions for LLM link."""

from copy import deepcopy

from fastmcp.tools import Tool as FastMCPTool
from openai.types.chat import ChatCompletionToolParam
from openai.types.shared_params.function_definition import FunctionDefinition


def transform_fastmcp_tool_to_openai_tool(fastmcp_tool: FastMCPTool) -> ChatCompletionToolParam:
    """Convert an FastMCP tool to an OpenAI tool."""

    tool_name = fastmcp_tool.name
    tool_description = fastmcp_tool.description or ""
    tool_parameters = deepcopy(fastmcp_tool.parameters)

    return ChatCompletionToolParam(
        type="function",
        function=FunctionDefinition(
            name=tool_name,
            description=tool_description,
            parameters=tool_parameters,
            strict=False,
        ),
    )



"""Sample code for FastMCP using InterceptingProxyTool."""

import asyncio
from typing import Any

from fastmcp import FastMCP
from fastmcp.client import Client
from fastmcp_agents.vendored.tool_transformer.tool_transformer import proxy_tool
from fastmcp.exceptions import ToolError

third_party_mcp_config = {
    "time": {
        "command": "uvx",
        "args": [
            "git+https://github.com/modelcontextprotocol/servers.git@2025.4.24#subdirectory=src/time",
            "--local-timezone=America/New_York",
        ],
    }
}


async def async_main():
    async with Client(third_party_mcp_config) as remote_mcp_client:
        proxied_mcp_server = FastMCP.as_proxy(remote_mcp_client)

        proxied_tools = await proxied_mcp_server.get_tools()

        frontend_server = FastMCP("Frontend Server")

        async def pre_call_hook(
            tool_args: dict[str, Any],
            hook_args: dict[str, Any],
        ) -> None:
            if tool_args.get("source_timezone") == "America/New_York":
                raise ToolError("New Yorkers are not allowed to use this tool.")

        proxy_tool(
            proxied_tools["convert_time"],
            frontend_server,
            name="transformed_convert_time",
            description="Converts a time from New York to another timezone.",
            pre_call_hook=pre_call_hook,
        )

        await frontend_server.run_async(transport="sse")


def run_mcp():
    asyncio.run(async_main())


if __name__ == "__main__":
    run_mcp()



import contextlib
import os
import tempfile
from collections.abc import AsyncGenerator, Callable, Coroutine
from functools import wraps
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, MagicMock

import pytest
from fastmcp import Client, FastMCP
from fastmcp.server.proxy import FastMCPProxy
from fastmcp.tools import Tool as FastMCPTool
from mcp.types import EmbeddedResource, ImageContent, TextContent

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.bundled.evaluator_optimizer import EvaluationResult, evaluate_conversation_factory, evaluate_result_factory
from fastmcp_agents.cli.loader import get_config_for_bundled
from fastmcp_agents.cli.models import AugmentedServerModel, OverriddenStdioMCPServer, ServerSettings
from fastmcp_agents.conversation.types import (
    CallToolResponse,
)


class ReturnTrackingAsyncMock(AsyncMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.call_return_value_list = []

    async def _execute_mock_call(self, *args, **kwargs):
        value = await super()._execute_mock_call(*args, **kwargs)
        self.call_return_value_list.append(value)
        return value


@pytest.fixture
def server_config_name():
    """Override this fixture in test files to specify which config to use."""
    msg = "Override this fixture in test files to specify which config to use."
    raise NotImplementedError(msg)


@pytest.fixture
def server_settings():
    """Override this fixture in test files to specify server settings."""
    return ServerSettings(transport="stdio", log_level="DEBUG")


@pytest.fixture
def cli_server_config(server_config_name):
    """Get the server configuration based on the config name."""
    return get_config_for_bundled(server_config_name)


@pytest.fixture
async def compiled_configuration(
    cli_server_config: AugmentedServerModel, server_settings: ServerSettings
) -> AsyncGenerator[tuple[list[FastMCPAgent], list[Client], FastMCP], None]:
    """Compile the server configuration with the given settings."""
    agents, mcp_clients, server = await cli_server_config.to_fastmcp_server(server_settings=server_settings)

    try:
        yield agents, mcp_clients, server
    finally:
        for client in mcp_clients:
            await client.close()


@pytest.fixture
def fastmcp_server(compiled_configuration: tuple[list[FastMCPAgent], list[Client], FastMCP]) -> FastMCP:
    """Get the FastMCP server instance."""
    return compiled_configuration[2]


@pytest.fixture
def mcp_clients(compiled_configuration: tuple[list[FastMCPAgent], list[Client], FastMCP]) -> list[Client]:
    """Get the list of MCP clients."""
    return compiled_configuration[1]


@pytest.fixture
def agents(compiled_configuration: tuple[list[FastMCPAgent], list[Client], FastMCP]) -> list[FastMCPAgent]:
    """Get the list of FastMCP agents."""
    return compiled_configuration[0]


@pytest.fixture
def agent_name():
    msg = "Override this fixture in test files to specify the name of the agent you are testing."
    raise NotImplementedError(msg)


@pytest.fixture
def agent(agent_name: str, agents: list[FastMCPAgent]) -> FastMCPAgent:
    agent = next(agent for agent in agents if agent.name == agent_name)
    agent.run = ReturnTrackingAsyncMock(wraps=agent.run)
    agent.call_tool = ReturnTrackingAsyncMock(wraps=agent.call_tool)
    return agent


@pytest.fixture
def agent_tool_calls(agent: FastMCPAgent) -> list[CallToolResponse]:
    assert isinstance(agent.call_tool, ReturnTrackingAsyncMock)

    return agent.call_tool.call_return_value_list


@pytest.fixture
def fastmcp_server_client(fastmcp_server: FastMCP) -> Client:
    """Create a client for the FastMCP server."""
    return Client(transport=fastmcp_server, timeout=120, init_timeout=30)


@pytest.fixture
async def initialized_client(fastmcp_server_client: Client) -> AsyncGenerator[Client, None]:
    """Initialize and yield a client, ensuring proper cleanup."""
    async with fastmcp_server_client:
        try:
            yield fastmcp_server_client
        finally:
            with contextlib.suppress(Exception):
                await fastmcp_server_client._disconnect()
                await fastmcp_server_client.close()


@pytest.fixture
def call_curator(
    initialized_client: Client,
) -> Callable[[str, str], Coroutine[Any, Any, list[TextContent | ImageContent | EmbeddedResource]]]:
    async def call_curator(name: str, instructions: str) -> list[TextContent | ImageContent | EmbeddedResource]:
        return await initialized_client.call_tool(name=name, arguments={"instructions": instructions})

    return call_curator


@pytest.fixture
def proxy_server(fastmcp_server_client: Client) -> FastMCPProxy:
    """Create a proxy server from the client."""
    return FastMCP.as_proxy(fastmcp_server_client)


@pytest.fixture
async def temp_working_dir() -> AsyncGenerator[Path, None]:
    original_dir = Path.cwd()

    with tempfile.TemporaryDirectory() as temp_dir:
        os.chdir(temp_dir)

        yield Path(temp_dir)

        os.chdir(original_dir)


@pytest.fixture
async def temp_dir() -> AsyncGenerator[Path, None]:
    with tempfile.TemporaryDirectory() as temp_dir:
        yield Path(temp_dir)


@pytest.fixture
async def evaluation_criteria() -> str | None:
    return None


@pytest.fixture
async def evaluator_agent(evaluation_criteria: str | None, server_settings: ServerSettings) -> FastMCPTool:
    config = get_config_for_bundled("evaluator_optimizer")

    optimizer = config.mcpServers["evaluator_optimizer"]
    assert isinstance(optimizer, OverriddenStdioMCPServer)
    if evaluation_criteria:
        optimizer.env = {**optimizer.env, "EVALUATION_CRITERIA": evaluation_criteria}

    _, _, server = await config.to_fastmcp_server(server_settings=server_settings)

    server_tools = await server.get_tools()

    return server_tools["evaluate_result"]


@pytest.fixture
def evaluator() -> Callable[..., Coroutine[Any, Any, EvaluationResult]]:
    async def evaluate(criteria: str, goal: str, proposed_solution: str) -> EvaluationResult:
        evaluate_function = evaluate_result_factory(criteria=criteria)

        return await evaluate_function(ctx=MagicMock(), goal=goal, proposed_solution=proposed_solution)

    return evaluate


def evaluate_with_criteria(criteria: str, minimum_grade: float = 0.9):
    """Decorator to evaluate test results against specified criteria.

    Args:
        criteria: The evaluation criteria to check against
        expected_grade: The expected grade (defaults to "A")
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            # Run the original test function
            agent, goal, solution = await func(*args, **kwargs)

            # Extract the conversation from the return value
            conversation = agent.run.call_return_value_list[0][0]

            # Create evaluator directly
            evaluate_function = evaluate_conversation_factory(criteria=criteria)

            # Run the evaluation
            evaluation = await evaluate_function(ctx=MagicMock(), goal=goal, proposed_solution=solution, conversation=conversation)

            evaluation_grade: float = evaluation.grade  # type: ignore

            # Assert the grade matches expectations
            assert evaluation_grade >= minimum_grade, (
                f"Expected grade {minimum_grade}, got {evaluation_grade}. Feedback: {evaluation.feedback}"
            )

            return goal, solution, evaluation

        return wrapper

    return decorator



"""Tools for redirecting tool calls."""

import json
from copy import deepcopy
from pathlib import Path
from typing import Any

from fastmcp import Context
from mcp.types import EmbeddedResource, ImageContent, TextContent


async def redirect_to_tool_call(
    ctx: Context,
    first_tool_name: str,
    first_tool_arguments: dict[str, Any],
    second_tool_name: str,
    second_tool_arguments: dict[str, Any],
    second_tool_response_argument: str,
) -> list[TextContent | ImageContent | EmbeddedResource]:
    """Redirect the result of any existing tool call to another tool call.

    Args:
        first_tool_name: The name of the first tool to call.
        first_tool_arguments: The arguments to pass to the first tool.
        second_tool_name: The name of the second tool to call.
        second_tool_arguments: The arguments to pass to the second tool.
        second_tool_response_argument: The name of the argument to insert the result of the first tool call into.

    Returns:
        A list of content from the second tool call.
    """

    tools = await ctx.fastmcp.get_tools()
    first_tool = tools[first_tool_name]
    second_tool = tools[second_tool_name]

    first_tool_result = await first_tool.run(first_tool_arguments)

    second_tool_full_arguments = deepcopy(second_tool_arguments)
    second_tool_full_arguments[second_tool_response_argument] = first_tool_result
    return await second_tool.run(second_tool_full_arguments)


async def redirect_to_file(ctx: Context, tool_name: str, arguments: dict, path: Path) -> bool:
    """Redirect any tool call to a local file.

    This is useful for debugging tools that are not available in the local environment.

    Args:
        tool_name: The name of the tool to redirect.
        arguments: The arguments to pass to the tool.
        path: The path to the file to write the tool call result to.

    Returns:
        True if the tool call was successful, False otherwise.
    """

    tools = await ctx.fastmcp.get_tools()
    tool = tools[tool_name]

    tool_call_result = await tool.run(arguments)

    with Path(path).open("w", encoding="utf-8") as f:
        f.write(json.dumps(tool_call_result, indent=2))

    return True



"""Ephemeral memory implementation."""

from fastmcp_agents.conversation.memory.base import MemoryProtocol
from fastmcp_agents.conversation.types import Conversation, ConversationEntryTypes


class EphemeralMemory(MemoryProtocol):
    """A memory entry that is lost when the server restarts"""

    conversation_history: Conversation

    def __init__(self):
        self.conversation_history = Conversation()

    def add(self, message: ConversationEntryTypes):
        self.conversation_history.append(message)

    def get(self):
        return self.conversation_history.get()

    def set(self, conversation_history: Conversation):
        self.conversation_history = conversation_history

    def reset(self) -> None:
        pass

    def persist(self) -> None:
        pass

    def restore(self) -> None:
        pass



"""Utility functions for conversation handling."""

from mcp.types import BlobResourceContents, EmbeddedResource, ImageContent, TextContent, TextResourceContents


def join_content(content: list[TextContent | ImageContent | EmbeddedResource]) -> str:
    """Join the content of a list of TextContent, ImageContent, or EmbeddedResource into a single string.

    Args:
        content: The list of content to join.

    Returns:
        A string of the joined content.
    """

    result = ""

    for item in content:
        if isinstance(item, TextContent):
            result += item.text
        elif isinstance(item, ImageContent):
            result += f"an image {item.mimeType}"
        elif isinstance(item, BlobResourceContents):
            result += f"a blob {item.mimeType}"
        elif isinstance(item, TextResourceContents):
            result += f"a text resource {item.text}"
        elif isinstance(item, EmbeddedResource):
            result += f"an embedded resource {item.type}"

    return result



import json
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from fastmcp import FastMCP

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.types import TextContent
from tests.conftest import evaluate_with_criteria

if TYPE_CHECKING:
    from fastmcp.tools import Tool as FastMCPTool


@pytest.fixture
def server_config_name():
    return "motherduckdb_mcp-server-motherduck"


class TestDuckDBAgent:
    @pytest.fixture
    def agent_name(self):
        return "ask_duckdb_agent"

    @pytest.fixture
    async def populate_database(self, fastmcp_server: FastMCP):
        tools = await fastmcp_server.get_tools()

        query_tool: FastMCPTool = tools["query"]

        await query_tool.run(
            arguments={
                "query": """
            CREATE TABLE people (id INTEGER, name TEXT);
            INSERT INTO people (id, name) VALUES (1, 'John');
            INSERT INTO people (id, name) VALUES (2, 'Jane');
            INSERT INTO people (id, name) VALUES (3, 'Jim');
            INSERT INTO people (id, name) VALUES (4, 'Jill');
            INSERT INTO people (id, name) VALUES (5, 'Jack');
            """
            }
        )

        await query_tool.run(
            arguments={
                "query": """
            CREATE TABLE dogs (id INTEGER, name TEXT, age INTEGER, color TEXT, sibling_ages INTEGER[]);
            INSERT INTO dogs (id, name, age, color, sibling_ages) VALUES (1, 'Fido', 100, 'red', [1, 2, 3]);
            INSERT INTO dogs (id, name, age, color, sibling_ages) VALUES (2, 'Buddy', 200, 'blue', [3, 4, 8]);
            INSERT INTO dogs (id, name, age, color, sibling_ages) VALUES (3, 'Max', 300, 'green', [7, 8, 9]);
            INSERT INTO dogs (id, name, age, color, sibling_ages) VALUES (4, 'Bella', 400, 'purple', [9, 10, 11]);
            INSERT INTO dogs (id, name, age, color, sibling_ages) VALUES (5, 'Lucy', 500, 'yellow', [13, 1, 15]);
            """
            }
        )

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the current tables were listed
        2. that the schema of each table was shown
        3. that a sample of data from each table was displayed
        4. that the agent used the correct sequence of tools
        5. that the agent did not run any unnecessary queries

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_database_inspection(
        self, temp_working_dir: Path, populate_database, agent: FastMCPAgent, call_curator, agent_tool_calls
    ):
        instructions = """
        Show me the current tables in the database, their schemas, and a sample of data from each table.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify database inspection was performed
        assert "table" in text_result.lower()
        assert "schema" in text_result.lower()
        assert "sample" in text_result.lower()

        # Verify tool calls
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert len(tool_call_names) > 4

        assert "tips_writing_queries" in tool_call_names
        assert "query" in tool_call_names

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the JSON data was successfully loaded
        2. that the table was created with the correct schema
        3. that the data was properly imported
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_json_loading(self, temp_working_dir: Path, agent: FastMCPAgent, fastmcp_server: FastMCP, call_curator, agent_tool_calls):
        instructions = """
        Load the JSON data from the file 'people.json' into a table called 'people'.
        Load the JSON data from the file 'dogs.json' into a table called 'dogs'.
        Show me the schema and a sample of the data from both tables.
        """

        # Create a temporary file with JSON data
        people_json = [
            {"id": 1, "name": "Tom"},
            {"id": 2, "name": "Dick"},
            {"id": 3, "name": "Harry"},
        ]
        dogs_json = [
            {"id": 1, "name": "Fido", "age": 100, "color": "red", "sibling_ages": [1, 2, 3]},
            {"id": 2, "name": "Buddy", "age": 200, "color": "blue", "sibling_ages": [3, 4, 8]},
            {"id": 3, "name": "Kepler", "age": 300, "color": "green", "sibling_ages": [7, 8, 9]},
            {"id": 4, "name": "Bella", "age": 400, "color": "purple", "sibling_ages": [9, 10, 11]},
            {"id": 5, "name": "Lucy", "age": 500, "color": "yellow", "sibling_ages": [13, 1, 15]},
        ]

        with Path("people.json").open("w", encoding="utf-8") as f:
            json.dump(people_json, f)
        with Path("dogs.json").open("w", encoding="utf-8") as f:
            json.dump(dogs_json, f)

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify JSON was loaded
        assert "json" in text_result.lower()

        # Verify tool calls
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert len(tool_call_names) > 4
        assert "query" in tool_call_names
        assert "tips_load_json" in tool_call_names or "tips_load_json_file" in tool_call_names

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the query was successfully executed
        2. that the results were properly formatted
        3. that the query logic was explained
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_query_execution(
        self, temp_working_dir: Path, populate_database, agent: FastMCPAgent, fastmcp_server: FastMCP, call_curator, agent_tool_calls
    ):
        instructions = """
        Execute the following query and explain the results:
        SELECT * FROM my_data WHERE age > 100 ORDER BY birthyear DESC LIMIT 5;
        """
        tools = await fastmcp_server.get_tools()

        query_tool: FastMCPTool = tools["query"]

        result = await query_tool.run(
            arguments={
                "query": """
            CREATE TABLE my_data (id INTEGER, name TEXT, birthyear INTEGER, age INTEGER, color TEXT, sibling_ages INTEGER[]);
            INSERT INTO my_data (id, name, birthyear, age, color, sibling_ages) VALUES (1, 'Fido', 1900, 100, 'red', [1, 2, 3]);
            INSERT INTO my_data (id, name, birthyear, age, color, sibling_ages) VALUES (2, 'Buddy', 1800, 200, 'blue', [4, 5, 6]);
            INSERT INTO my_data (id, name, birthyear, age, color, sibling_ages) VALUES (3, 'Max', 1700, 300, 'green', [7, 8, 9]);
            INSERT INTO my_data (id, name, birthyear, age, color, sibling_ages) VALUES (4, 'Bella', 1600, 400, 'purple', [10, 11, 12]);
            INSERT INTO my_data (id, name, birthyear, age, color, sibling_ages) VALUES (5, 'Lucy', 1500, 500, 'yellow', [13, 14, 15]);
            """
            }
        )
        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify query was executed
        assert "query" in text_result.lower()
        assert "results" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 1
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]

        assert "query" in tool_call_names

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the visualization was created
        2. that the data was properly formatted for visualization
        3. that the visualization type was appropriate
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_data_visualization(self, temp_working_dir: Path, populate_database, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        Create an ascii visualization of the data in dogs table.
        Show the distribution of values in sibling ages.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify visualization was created
        assert "distribution" in text_result.lower()
        assert "age" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 1
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert "tips_writing_queries" in tool_call_names
        assert "query" in tool_call_names

        return agent, instructions, text_result



"""Custom errors for LLM link."""

from fastmcp_agents.errors.base import FastMCPAgentsError


class ModelDoesNotSupportFunctionCallingError(FastMCPAgentsError):
    """Raised when a model does not support function calling."""

    def __init__(self, model: str):
        self.message = f"Model {model} does not support function calling"
        super().__init__(self.message)


class ModelDoesNotExistError(FastMCPAgentsError):
    """Raised when a model does not exist."""

    def __init__(self, model: str):
        self.message = f"Model {model} does not exist"
        super().__init__(self.message)


class ModelNotSetError(FastMCPAgentsError):
    """Raised when a model is not set."""

    def __init__(self):
        self.message = "MODEL environment variable is not set and no model was provided"
        super().__init__(self.message)



"""Sample code for FastMCP using InterceptingProxyTool."""

import asyncio
from typing import TYPE_CHECKING

from fastmcp.client import Client

if TYPE_CHECKING:
    from fastmcp.tools import Tool as FastMCPTool

third_party_mcp_config = {
    "time": {
        "command": "uvx",
        "args": [
            "git+https://github.com/modelcontextprotocol/servers.git@2025.4.24#subdirectory=src/time",
            "--local-timezone=America/New_York",
        ],
    }
}


async def async_main():
    client = Client(third_party_mcp_config, timeout=30)

    client_tools = await client.list_tools()
    client_tools_by_name = {tool.name: tool for tool in client_tools}

    convert_time_tool: FastMCPTool = client_tools_by_name["convert_time"]

    convert_time_tool.run(arguments={"source_timezone": "America/New_York", "time": "3:00", "target_timezone": "America/Los_Angeles"})


def run_mcp():
    asyncio.run(async_main())


if __name__ == "__main__":
    run_mcp()



"""Tools for evaluating and optimizing task results."""

import asyncio
from collections.abc import Callable, Coroutine
from typing import Any, Literal

import asyncclick as click
import yaml
from fastmcp import Context, FastMCP
from fastmcp.tools import Tool as FastMCPTool
from pydantic import BaseModel, Field, computed_field

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.types import Conversation

evaluator = FastMCPAgent(
    name="task_result_evaluator",
    description="Evaluates the result of a task and provides feedback on the quality of the result.",
    system_prompt="You are a helpful assistant that evaluates the result of a task and provides feedback on the quality of the result.",
)

A: float = 0.9
B: float = 0.8
C: float = 0.7
D: float = 0.6
F: float = 0.0


class CriteriaNotes(BaseModel):
    """A judging criteria with notes."""

    criteria: str = Field(..., description="The description of the criteria.")
    notes: str = Field(..., description="Any notes on the criteria.")


class CriteriaScore(CriteriaNotes):
    """A judging criteria with a score and notes."""

    max_points: int = Field(..., description="The maximum points of the score part.", ge=0)
    points: int = Field(..., description="The points of the score part.", ge=0)


class EvaluationResult(BaseModel):
    """A result of an evaluation."""

    criteria: list[CriteriaScore] = Field(..., description="The criteria for the evaluation.")

    @property
    def feedback(self) -> dict[str, str]:
        """Get the feedback for the evaluation."""
        return {criteria.criteria: criteria.notes for criteria in self.criteria if criteria.notes}

    @computed_field(return_type=float)
    def grade(self):
        """Get the grade for the evaluation."""
        return self._grade()

    @computed_field
    def letter_grade(self) -> Literal["A", "B", "C", "D", "F"]:
        """Get the letter grade for the evaluation."""
        grade: float = self._grade()
        if grade >= A:
            return "A"
        if grade >= B:
            return "B"
        if grade >= C:
            return "C"
        if grade >= D:
            return "D"
        return "F"

    def _grade(self):
        """Get the grade for the evaluation."""
        # Pylance was being weird so I moved it into a private function
        if not self.criteria:
            return 0.0
        return sum(criteria.points for criteria in self.criteria) / sum(criteria.max_points for criteria in self.criteria)


class EvaluationError(Exception):
    """An error that occurs during evaluation."""

    def __init__(self, message: str):
        """Initialize the error."""
        super().__init__(message)


def build_prompt(goal: str, proposed_solution: str, criteria: str, conversation_history: Conversation | None = None) -> str:
    """Build the prompt for the evaluation."""
    prompt = EVALUATION_PREAMBLE

    format_kwargs = {
        "goal": goal,
        "proposed_solution": proposed_solution,
        "criteria": criteria,
    }

    if conversation_history:
        prompt += CONVERSATION_SUFFIX
        format_kwargs["conversation_history"] = yaml.safe_dump(
            compress_messages(conversation_history.to_messages()), indent=2, sort_keys=True
        )

    return prompt.format(**format_kwargs)


def compress_message(message: dict[str, Any]) -> dict[str, Any]:
    """Compress a message to 1KB."""
    for key, value in message.items():
        if isinstance(value, str) and len(value) > ONE_KB:
            message[key] = value[:ONE_KB] + "..."
    return message


def compress_messages(messages: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """Compress a list of messages to 1KB."""
    return [compress_message(message) for message in messages]


ONE_KB = 1024


async def perform_evaluation(ctx: Context, prompt: str) -> EvaluationResult:
    _, evaluation_result = await evaluator.run(
        ctx,
        instructions=prompt,
        success_response_model=EvaluationResult,
        raise_on_error_response=True,
    )

    if not isinstance(evaluation_result, EvaluationResult):
        raise EvaluationError(message="The evaluation result is not a EvaluationResult")

    return evaluation_result


async def evaluate_conversation(
    ctx: Context, criteria: str, goal: str, proposed_solution: str, conversation: Conversation
) -> EvaluationResult:
    prompt = build_prompt(
        goal=goal,
        proposed_solution=proposed_solution,
        criteria=criteria,
        conversation_history=conversation,
    )

    return await perform_evaluation(ctx, prompt)


async def evaluate_result(ctx: Context, criteria: str, goal: str, proposed_solution: str) -> EvaluationResult:
    """Evaluate the result of a task."""
    prompt = build_prompt(goal=goal, proposed_solution=proposed_solution, criteria=criteria)

    return await perform_evaluation(ctx, prompt)


def evaluate_conversation_factory(criteria: str) -> Callable[..., Coroutine[Any, Any, EvaluationResult]]:
    """A factory for creating evaluation functions with a fixed criteria."""

    async def evaluate(ctx: Context, goal: str, proposed_solution: str, conversation: Conversation) -> EvaluationResult:
        return await evaluate_conversation(
            ctx=ctx, criteria=criteria, goal=goal, proposed_solution=proposed_solution, conversation=conversation
        )

    return evaluate


def evaluate_result_factory(criteria: str) -> Callable[..., Coroutine[Any, Any, EvaluationResult]]:
    """A factory for creating evaluation functions with a fixed criteria."""

    async def evaluate(ctx: Context, goal: str, proposed_solution: str) -> EvaluationResult:
        return await evaluate_result(ctx=ctx, criteria=criteria, goal=goal, proposed_solution=proposed_solution)

    return evaluate


@click.command()
@click.option(
    "--evaluation-criteria",
    type=str,
    envvar="evaluation_criteria",
    help="The evaluation criteria to use for scoring the result.",
)
async def cli(evaluation_criteria: str | None = None):
    evaluator_tool = FastMCPTool.from_function(fn=evaluate_result_factory(evaluation_criteria or DEFAULT_CRITERIA))

    mcp = FastMCP(name="Local Evaluate Optimize", tools=[evaluator_tool])

    await mcp.run_async()


def run_mcp():
    asyncio.run(cli())


if __name__ == "__main__":
    run_mcp()

DEFAULT_CRITERIA = """
| Criteria | Description | Points |
|----------|-------------|---------|
| Completeness | The proposed solution is complete, relevant, and covers all the aspects of the goal. | 10 |
| Accuracy | The proposed solution is accurate and correct. | 10 |
| Simplicity | The proposed solution is the simplest answer that totally achieves the stated goal. | 10 |
| Clarity | The proposed solution is clear and easy to understand. | 10 |
"""

EVALUATION_PREAMBLE = """
You are a helpful assistant that evaluates the final work product of someone who has been working to achieve a goal.

You will not do any of the work yourself, you are only evaluating the final work product. You are an objective observer
who is not swayed by errors encountered, problems, etc. You only care whether the work product achieves the goal.

You must provide a score between 0 and 100 for the result. You must provide a feedback on the result.

## Illustrative Example

Imagine you are judging a competition.

The task is to: `Write a Python function that calculates the square of a number.`

| Criteria | Description | Points |
|----------|-------------|---------|
| Completeness | The proposed solution is complete, relevant, and covers all the aspects of the goal. | 10 |
| Accuracy | The proposed solution is accurate and correct. | 10 |
| Simplicity | The proposed solution is the simplest answer that totally achieves the stated goal. | 10 |
| Clarity | The proposed solution is clear and easy to understand. | 10 |


The proposed solution is:
```python
def multiply(x, y):
    if not isinstance(x, int) or not isinstance(y, int):
        raise ValueError("x and y must be integers")
    return x * y

def square(x):
    first_x = x
    second_x = x
    return multiply(first_x, second_x)
```

You thoroughly evaluate the proposed solution and provide a score and feedback on the result.

| Criteria | Score | Notes |
|----------|-------|-------|
| Completeness | 10 | None |
| Accuracy | 10 | None |
| Simplicity | 4 | A simpler solution would be to use the `**` operator. |
| Clarity | 9 | The code would be more clear if it was commented. |

Total Score: 33 out of 40 (82.5%)

# The Evaluation

The goal of the task is:
```
{goal}
```

The proposed solution is:
```
{proposed_solution}
```

The evaluation criteria is:
```markdown
{criteria}
```
"""

CONVERSATION_SUFFIX = """
The conversation history is:
```
{conversation_history}
```

Note: This is a worklog from the agent that was working to achieve the goal. Entries longer than 1000 characters have
been truncated and will end with "..." to indicate that they have been truncated.

You can check the worklog to make sure that the agent:
1) did not miss any important information
2) did not make any mistakes
3) did not invent a positive result that was not actually achieved
"""



"""Pydantic models for conversation and tool calls."""

import json
from typing import Any, Literal, TypeAlias

from mcp.types import EmbeddedResource, ImageContent, TextContent
from openai.types.chat.chat_completion_message_tool_call_param import ChatCompletionMessageToolCallParam
from pydantic import BaseModel, Field, model_serializer

MCPToolResponseTypes = TextContent | ImageContent | EmbeddedResource


class CallToolRequest(BaseModel):
    """A helper class for a tool call request."""

    id: str = Field(
        ..., description="The id of the tool call request. This is used to match the tool call request to the tool call response."
    )
    name: str = Field(..., description="The name of the tool to call.")
    arguments: dict[str, Any] = Field(..., description="The arguments to pass to the tool.")

    @classmethod
    def from_openai(cls, message: ChatCompletionMessageToolCallParam) -> "CallToolRequest":
        function = message["function"]
        return cls(id=message["id"], name=function["name"], arguments=json.loads(function["arguments"]))

    @model_serializer
    def serialize(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "function": {
                "name": self.name,
                "arguments": json.dumps(self.arguments),
            },
            "type": "function",
        }


class CallToolResponse(BaseModel):
    """A helper class for a tool call response."""

    id: str = Field(
        ..., description="The id of the tool call request. This is used to match the tool call request to the tool call response."
    )
    name: str = Field(..., description="The name of the tool to call.")
    arguments: dict[str, Any] = Field(..., description="The arguments that were passed to the tool.", exclude=True)
    content: list[TextContent | ImageContent | EmbeddedResource] = Field(..., description="The content of the tool call response.")


class SystemConversationEntry(BaseModel):
    """A chat entry is a message in the chat history."""

    role: Literal["system"] = Field(default="system", description="A conversation entry that is a system message")
    content: str = Field(..., description="The content of the chat entry")


class UserConversationEntry(BaseModel):
    """A chat entry is a message in the chat history."""

    role: Literal["user"] = Field(default="user", description="A conversation entry that is a user message")
    content: str = Field(..., description="The content of the chat entry")


class ToolConversationEntry(BaseModel):
    """A tool call chat entry is a message in the chat history that is a tool call."""

    role: Literal["tool"] = Field(default="tool", description="A conversation entry that is a tool call")
    tool_call_id: str = Field(..., description="The id of the tool call request.")
    name: str = Field(..., description="The name of the tool to call.")
    content: list[TextContent | ImageContent | EmbeddedResource] = Field(..., description="The content of the tool call response.")

    @classmethod
    def from_tool_call_response(cls, tool_call_response: CallToolResponse) -> "ToolConversationEntry":
        return cls(
            role="tool",
            tool_call_id=tool_call_response.id,
            name=tool_call_response.name,
            content=tool_call_response.content,
        )


class AssistantConversationEntry(BaseModel):
    """A chat entry is a message in the chat history."""

    role: Literal["assistant"] = Field(default="assistant", description="A conversation entry that is an assistant message")
    content: str | None = Field(default=None, description="The content of the chat entry")
    tool_calls: list[CallToolRequest] = Field(default_factory=list, description="The tool calls that were made in the assistant message")
    token_usage: int | None = Field(default=None, description="The number of tokens used by the assistant message")


ConversationEntryTypes: TypeAlias = SystemConversationEntry | UserConversationEntry | AssistantConversationEntry | ToolConversationEntry


class Conversation(BaseModel):
    entries: list[ConversationEntryTypes] = Field(default_factory=list[ConversationEntryTypes], frozen=True)

    def append(self, message: ConversationEntryTypes) -> "Conversation":
        """Returns a new conversation with the message appended."""
        return self.model_copy(update={"entries": [*self.entries, message]})

    def extend(self, entries: list[ConversationEntryTypes]) -> "Conversation":
        """Returns a new conversation with the messages extended."""
        return self.model_copy(update={"entries": [*self.entries, *entries]})

    def get(self) -> list[ConversationEntryTypes]:
        """Get the conversation history."""
        return self.entries

    def set(self, entries: list[ConversationEntryTypes]) -> "Conversation":
        """Returns a new conversation with the conversation history set."""
        return self.model_copy(update={"entries": entries})

    def to_messages(self) -> list[dict[str, Any]]:
        """Convert the conversation to a list of dictionaries."""
        return [message.model_dump() for message in self.entries]

    def merge(self, conversation: "Conversation") -> "Conversation":
        """Returns a new conversation with the conversation history merged."""
        return self.model_copy(update={"entries": [*self.entries, *conversation.entries]})



"""Sample code for FastMCP using InterceptingProxyTool."""

import asyncio
from pathlib import Path

import yaml

from fastmcp import FastMCP
from fastmcp.client import Client
from fastmcp_agents.vendored.tool_transformer.loader import overrides_from_yaml_file
from fastmcp_agents.vendored.tool_transformer.tool_transformer import proxy_tool

third_party_mcp_config = {
    "time": {
        "command": "uvx",
        "args": [
            "git+https://github.com/modelcontextprotocol/servers.git@2025.4.24#subdirectory=src/time",
            "--local-timezone=America/New_York",
        ],
    }
}

override_config_yaml = yaml.safe_load("""
tools:
- convert_time:
    description: >-
        An updated multi-line description 
        for the time tool.
    parameter_overrides:
        source_timezone:
        description: This field now has a description and a constant value
        constant: America/New_York
        time:
        description: This field now has a description and a default value
        default: "3:00"
""")


async def async_main():
    async with Client(third_party_mcp_config) as remote_mcp_client:
        backend_server = FastMCP.as_proxy(remote_mcp_client)

        backend_tools = await backend_server.get_tools()

        frontend_server = FastMCP("Frontend Server")

        tool_overrides = overrides_from_yaml_file(Path("override_config.yaml"))

        proxy_tool(
            tool=backend_tools["convert_time"],
            server=frontend_server,
            override=tool_overrides["convert_time"],
        )

        await frontend_server.run_async(transport="sse")


def run_mcp():
    asyncio.run(async_main())


if __name__ == "__main__":
    run_mcp()



"""Functions for loading configurations."""

from pathlib import Path
from urllib.parse import ParseResult, urlparse

import requests
import yaml

from fastmcp_agents.cli.models import (
    AugmentedServerModel,
    OverriddenStdioMCPServer,
)
from fastmcp_agents.observability.logging import BASE_LOGGER

logger = BASE_LOGGER.getChild("loader")

BUNDLED_DIR = Path(__file__).parent.parent / "bundled"

DEFAULT_ARG_REPLACEMENTS: dict[str, str] = {
    "bundled": str(BUNDLED_DIR),
}


def get_config_from_string(config_string: str) -> AugmentedServerModel:
    """Serialize a config from a string."""
    return AugmentedServerModel.model_validate(yaml.safe_load(config_string))


def get_config_from_url(config_url: str) -> AugmentedServerModel:
    """Serialize a config from a URL."""
    url: ParseResult = urlparse(config_url)

    config_raw = requests.get(url.geturl(), timeout=10).text
    config = get_config_from_string(config_raw)

    # If there is a file name in the url, remove it and use that as the relative path
    url_path_without_file = url.path.rsplit("/", 1)[0]

    replacements = {
        **DEFAULT_ARG_REPLACEMENTS,
        # Remove the path off the url
        "relative_to_this": url_path_without_file,
    }

    return process_arg_replacements(config, replacements)


def get_config_from_file(file: str, directory: str | None = None) -> AugmentedServerModel:
    """Serialize a config from a file."""

    file_path: Path = (Path(directory) / file) if directory else Path(file)

    if not file_path.exists():
        msg = f"Config file {file_path} not found"
        raise FileNotFoundError(msg)
    config_raw = file_path.read_text(encoding="utf-8")

    config = get_config_from_string(config_raw)

    replacements = {
        **DEFAULT_ARG_REPLACEMENTS,
        "relative_to_this": str(file_path.relative_to(Path(file_path).parent)),
    }

    return process_arg_replacements(config, replacements)


def get_config_for_bundled(config_bundled: str) -> AugmentedServerModel:
    """Serialize a config for a bundled server."""
    bundled_config_yml = BUNDLED_DIR / f"{config_bundled}.yml"

    if bundled_config_yml.exists():
        bundled_config_path = bundled_config_yml
        config_raw = bundled_config_path.read_text(encoding="utf-8")
        config = get_config_from_string(config_raw)
        return process_arg_replacements(config, DEFAULT_ARG_REPLACEMENTS)

    msg = f"Bundled server config file {bundled_config_yml} not found"
    raise FileNotFoundError(msg)


def process_arg_replacements(augmented_server_model: AugmentedServerModel, replacements: dict[str, str]) -> AugmentedServerModel:
    """Process arg replacements in a config string."""

    if replacements:
        stdio_servers = [server for server in augmented_server_model.mcpServers.values() if isinstance(server, OverriddenStdioMCPServer)]
        for server in stdio_servers:
            args: list[str] = server.args
            args = [arg.format(**replacements) if isinstance(arg, str) else arg for arg in args]
            server.args = args

    return augmented_server_model



"""OpenTelemetry configuration for FastMCP Agents."""

import os

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from fastmcp_agents.observability.logging import BASE_LOGGER


def setup_otel():
    import logging

    from opentelemetry import trace
    from opentelemetry._logs import set_logger_provider
    from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
    from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
    from opentelemetry.sdk._logs.export import BatchLogRecordProcessor

    provider = TracerProvider()
    processor = BatchSpanProcessor(OTLPSpanExporter())
    provider.add_span_processor(processor)

    # Sets the global default tracer provider
    trace.set_tracer_provider(provider)

    logger_provider = LoggerProvider()
    set_logger_provider(logger_provider)

    exporter = OTLPLogExporter()
    logger_provider.add_log_record_processor(BatchLogRecordProcessor(exporter))
    handler = LoggingHandler(level=logging.NOTSET, logger_provider=logger_provider)

    # Attach OTLP handler to root logger
    BASE_LOGGER.addHandler(handler)


if os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT") is not None:
    setup_otel()



from pathlib import Path

import pytest

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.types import TextContent
from tests.conftest import evaluate_with_criteria


@pytest.fixture
def server_config_name():
    return "cyanheads_git-mcp-server"


class TestGitAgent:
    @pytest.fixture
    def agent_name(self):
        return "ask_git_agent"

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation historymust indicate:
        1. that the repository was successfully cloned
        2. that it has been cloned to a path that is appropriate for the task
        3. that the agent used the correct sequence of git commands

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_ask_git_for_clone(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = "Do a depth 1 clone of the repository https://github.com/modelcontextprotocol/servers"

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Check if there is now a servers directory
        assert (temp_working_dir / "servers").exists()

        # Ensure it has a README.md file
        assert (temp_working_dir / "servers" / "README.md").exists()

        assert "success" in text_result.lower()
        assert "servers" in text_result.lower()

        assert len(agent_tool_calls) < 4
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        # Agent may also call set_working_dir to set the working directory
        assert "git_clone" in tool_call_names
        assert "report_success" in tool_call_names

        clone_tool_call = next(tool_call for tool_call in agent_tool_calls if tool_call.name == "git_clone")
        assert clone_tool_call.arguments == {
            "repositoryUrl": "https://github.com/modelcontextprotocol/servers",
            "targetPath": "servers",
            "depth": 1,
        }
        assert agent_tool_calls[1].name == "report_success"

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that a new branch was successfully created
        2. that the branch was created from the correct base branch
        3. that the working directory was switched to the new branch
        4. that the branch name follows good naming conventions (feature/, bugfix/, etc.)
        5. that the agent used the correct sequence of git commands

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_branch_management(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        1. Clone the repository https://github.com/modelcontextprotocol/servers
        2. Create a new feature branch for issue #1234 from main
        3. Switch to the new branch
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify the branch was created
        assert (temp_working_dir / "servers").exists()
        assert "feature/issue-1234" in text_result.lower()

        # Verify tool calls
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert len(agent_tool_calls) >= 3
        assert "git_clone" in tool_call_names
        assert "git_branch" in tool_call_names
        assert "git_checkout" in tool_call_names

        clone_tool_call = next(tool_call for tool_call in agent_tool_calls if tool_call.name == "git_clone")
        assert clone_tool_call.arguments == {
            "repositoryUrl": "https://github.com/modelcontextprotocol/servers",
            "targetPath": "servers",
        }
        branch_tool_call = next(tool_call for tool_call in agent_tool_calls if tool_call.name == "git_branch")
        assert branch_tool_call.arguments in [
            {
                "branchName": "feature/issue-1234",
                "baseBranch": "main",
            },
            {
                "branchName": "feature/issue-1234",
                "mode": "create",
                "startPoint": "main",
            },
        ]

        checkout_tool_call = next(tool_call for tool_call in agent_tool_calls if tool_call.name == "git_checkout")
        assert checkout_tool_call.arguments == {
            "branchOrPath": "feature/issue-1234",
        }

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the remote was successfully added
        2. that the remote URL was correctly set
        3. that the remote was verified to be accessible
        4. that the remote was listed in the remote list
        5. that the agent used the correct sequence of git commands

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_remote_management(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        1. Clone the repository https://github.com/modelcontextprotocol/servers
        2. Add a new remote called 'upstream' pointing to https://github.com/modelcontextprotocol/servers.git
        3. Verify the remote was added correctly
        4. List all remotes to confirm
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify remote was added
        assert "upstream" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 3
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]

        assert "git_clone" in tool_call_names
        assert "git_remote" in tool_call_names

        remote_add_tool_call = next(
            tool_call for tool_call in agent_tool_calls if tool_call.name == "git_remote" and tool_call.arguments.get("mode") == "add"
        )
        assert remote_add_tool_call.arguments == {
            "name": "upstream",
            "url": "https://github.com/modelcontextprotocol/servers.git",
            "mode": "add",
        }

        show_tool_call = next(
            tool_call for tool_call in agent_tool_calls if tool_call.name == "git_remote" and tool_call.arguments.get("mode") == "show"
        )
        assert show_tool_call.arguments == {
            "name": "upstream",
            "mode": "show",
        }

        list_tool_call = next(
            tool_call for tool_call in agent_tool_calls if tool_call.name == "git_remote" and tool_call.arguments.get("mode") == "list"
        )
        assert list_tool_call.arguments == {
            "mode": "list",
        }

        return agent, instructions, text_result



"""Base class for single-step agents."""

import logging
from abc import ABC

import yaml
from fastmcp import Context
from fastmcp.tools import Tool as FastMCPTool
from mcp.types import TextContent

from fastmcp_agents.conversation.types import (
    AssistantConversationEntry,
    CallToolRequest,
    CallToolResponse,
    Conversation,
    MCPToolResponseTypes,
    ToolConversationEntry,
    UserConversationEntry,
)
from fastmcp_agents.conversation.utils import join_content
from fastmcp_agents.errors.agent import ToolNotFoundError
from fastmcp_agents.llm_link.base import AsyncLLMLink
from fastmcp_agents.observability.logging import BASE_LOGGER

logger = BASE_LOGGER


class SingleStepAgent(ABC):
    """A single-step agent, which can pick tools, calls them or do both in a single step."""

    name: str
    description: str
    llm_link: AsyncLLMLink
    default_tools: list[FastMCPTool]
    _logger: logging.Logger
    _tool_logger: logging.Logger
    _system_prompt: Conversation

    def __init__(
        self,
        *args,
        name: str,
        description: str,
        llm_link: AsyncLLMLink,
        system_prompt: Conversation,
        default_tools: list[FastMCPTool],
        **kwargs,
    ):
        """Initialize the single-step agent.

        Args:
            name: The name of the agent.
            description: The description of the agent.
            llm_link: The LLM link to use.
            system_prompt: The system prompt to use.
            default_tools: The default tools to use.
        """

        self.name = name
        self.description = description

        self._system_prompt = system_prompt

        self.default_tools = default_tools
        self.llm_link = llm_link

        self._logger = logger.getChild(f"{self.name}")
        self._tool_logger = self._logger.getChild("tool_calls")
        self.llm_link.logger = self._logger.getChild("llm_link")

        super().__init__(*args, **kwargs)

    async def call_tool(
        self,
        tool_call_request: CallToolRequest,
        fastmcp_tool: FastMCPTool,
    ) -> CallToolResponse:
        """Run a single tool call request with a single tool."""

        self._tool_logger.info(f"Calling tool {tool_call_request.name} with arguments {tool_call_request.arguments}")

        try:
            tool_response: list[MCPToolResponseTypes] = await fastmcp_tool.run(arguments=tool_call_request.arguments)
        except Exception as e:
            tool_response = [TextContent(type="text", text=f"Error calling tool {tool_call_request.name}: {e!s}")]

        full_response = join_content(tool_response)
        self._tool_logger.info(f"{tool_call_request.name} returned {len(full_response)} bytes: {str(full_response)[:200]}...")

        return CallToolResponse(
            id=tool_call_request.id,
            name=tool_call_request.name,
            arguments=tool_call_request.arguments,
            content=tool_response,
        )

    async def call_tools(
        self,
        tool_call_requests: list[CallToolRequest],
        fastmcp_tools: list[FastMCPTool],
    ) -> list[CallToolResponse]:
        """Run a list of tool call requests with a list of tools.

        Args:
            tool_call_requests: The tool call requests to run.
            fastmcp_tools: The tools to use.

        Returns:
            The tool call responses.
        """

        for tool_call_request in tool_call_requests:
            if tool_call_request.name not in [tool.name for tool in fastmcp_tools]:
                raise ToolNotFoundError(self.name, tool_call_request.name)

        return [
            await self.call_tool(tool_call_request, fastmcp_tool)
            for tool_call_request in tool_call_requests
            for fastmcp_tool in fastmcp_tools
            if tool_call_request.name == fastmcp_tool.name
        ]

    async def pick_tools(
        self,
        prompt: str | Conversation,
        tools: list[FastMCPTool],
    ) -> AssistantConversationEntry:
        """Send the prompt to the LLM and ask it what tool calls it wants to make.

        Args:
            prompt: The prompt to send to the LLM to solicit tool call requests
            tools: The tools to use.

        Returns:
            A list of CallToolRequest objects.
        """

        if isinstance(prompt, str):
            prompt = self._system_prompt.append(UserConversationEntry(content=prompt))

        assistant_conversation_entry = await self.llm_link.async_completion(conversation=prompt, fastmcp_tools=tools)

        tool_calls = assistant_conversation_entry.tool_calls
        tokens = assistant_conversation_entry.token_usage

        tool_call_yaml = yaml.safe_dump(assistant_conversation_entry.model_dump(exclude_none=True), indent=2, sort_keys=True)
        self._logger.info(f"Agent picked {len(tool_calls)} tool calls ({tokens} tokens): {tool_call_yaml}")

        return assistant_conversation_entry

    async def run_step(
        self,
        *args,  # noqa: ARG002
        ctx: Context,  # noqa: ARG002
        prompt: str | Conversation,
        tools: list[FastMCPTool],
        **kwargs,  # noqa: ARG002
    ) -> tuple[AssistantConversationEntry, list[ToolConversationEntry]]:
        """Run a single step of the agent.

        This method is called to run a single step of the agent. It will:
        - Ask the LLM what tool calls it wants to make
        - Call the tools

        Args:
            ctx: The context of the FastMCP Request.
            prompt: The prompt to send to the LLM to solicit tool call requests
            tools: The tools to use.

        Returns:
            A tuple containing the assistant conversation entry with the tool call requests and the tool
            conversation entries with the tool call responses.
        """

        if isinstance(prompt, str):
            prompt = self._system_prompt.append(UserConversationEntry(content=prompt))

        assistant_conversation_entry = await self.pick_tools(prompt, tools)

        return assistant_conversation_entry, [
            ToolConversationEntry.from_tool_call_response(tool_call_response)
            for tool_call_response in await self.call_tools(assistant_conversation_entry.tool_calls, tools)
        ]



import subprocess
from pathlib import Path

import pytest

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.types import TextContent
from tests.conftest import evaluate_with_criteria


@pytest.fixture
def server_config_name():
    return "wrale_mcp-server-tree-sitter"


@pytest.fixture
def project_in_dir(temp_working_dir: Path):
    # git clone the tree-sitter-server repo
    subprocess.run(["git", "clone", "https://github.com/wrale/mcp-server-tree-sitter.git", temp_working_dir / "test_project"], check=False)
    return temp_working_dir / "test_project"


class TestTreeSitterAgent:
    @pytest.fixture
    def agent_name(self):
        return "ask_tree_sitter_agent"

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the project was successfully registered
        2. that the project was analyzed
        3. that the agent used the correct sequence of tools
        4. that the agent did not call any unnecessary tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_project_registration(self, project_in_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        1. Register the project in the test_project directory
        2. Produce a basic analysis of the project
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify project was registered and analyzed
        assert "registered" in text_result.lower()
        assert "python" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 3
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert "register_project_tool" in tool_call_names
        assert "analyze_project" in tool_call_names

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the project was successfully registered
        2. that the search for the specified text was performed
        3. that the results were properly formatted and returned
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_text_search(self, project_in_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        1. Register the project in the test_project directory
        2. Search for the text 'def test_' in all files
        3. Return the results in a clear format
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify search was performed
        assert "found" in text_result.lower()
        assert "def test_" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 3
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert "register_project_tool" in tool_call_names
        assert "find_text" in tool_call_names

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the project was successfully registered
        2. that the symbol search was performed
        3. that the results were properly formatted and returned
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_symbol_search(self, project_in_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        1. Register the project in the test_project directory
        2. Analyze the project and find all function definitions in the project
        3. Return the results in a clear format
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify symbol search was performed
        assert "found" in text_result.lower()
        assert "function" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 3
        tool_call_names = [tool_call.name for tool_call in agent_tool_calls]
        assert "register_project_tool" in tool_call_names
        assert "analyze_project" in tool_call_names
        assert "run_query" in tool_call_names

        return agent, instructions, text_result



"""Functions for transforming tools."""

from typing import Annotated

from fastmcp.tools.tool import Tool as FastMCPTool
from fastmcp.utilities.logging import get_logger

from fastmcp_agents.vendored.tool_transformer.models import (
    PostToolCallHookProtocol,
    PreToolCallHookProtocol,
    ToolOverride,
    ToolParameterTypes,
)

logger = get_logger(__name__)


def transform_tool(
    tool: FastMCPTool,
    override: Annotated[ToolOverride, "Tool overrides for the tool."] | None = None,
    *,
    name: Annotated[str, "The name for the transformed tool."] | None = None,
    description: Annotated[str, "The description for the transformed tool."] | None = None,
    parameter_overrides: Annotated[list[ToolParameterTypes], "Parameter overrides for the tool."] | None = None,
    hook_parameters: Annotated[list[ToolParameterTypes], "Hook parameters for the tool."] | None = None,
    pre_call_hook: Annotated[
        PreToolCallHookProtocol | None,
        "A hook that is called before a tool call is made.",
    ] = None,
    post_call_hook: Annotated[
        PostToolCallHookProtocol | None,
        "A hook that is called after a tool call is made.",
    ] = None,
) -> FastMCPTool:
    """
    Transforms a tool with the given overrides.

    Args:
        tool: The tool to transform.
        override: The overrides to apply to the tool. If not provided, the name, description, parameter_overrides,
                  hook_parameters, pre_call_hook, and post_call_hook will be used.
        name: The name for the transformed tool.
        description: The description for the transformed tool.
        parameter_overrides: The parameter overrides to apply to the tool.
        hook_parameters: The hook parameters to apply to the tool.
        pre_call_hook: The pre-call hook to apply to the tool.
        post_call_hook: The post-call hook to apply to the tool.

    Returns:
        The transformed tool.

    Examples:
        >>> transform_tool(
        ...     tool=Tool(
        ...         name="my_tool",
        ...         description="This is my tool",
        ...         parameters={"arg1": "string"},
        ...     ),
        ...     name="my_tool_with_overrides",
        ...     description="This is my tool with overrides",
        ...     parameter_overrides=[
        ...         ToolParameter(
        ...             name="arg1",
        ...             description="now it has a description",
        ...             default="now it has a default value",
        ...         ),
        ...     ],
        ... )
    """

    if not override:
        override = ToolOverride(
            name=name,
            description=description,
            parameter_overrides=parameter_overrides or [],
            hook_parameters=hook_parameters or [],
            pre_call_hook=pre_call_hook,
            post_call_hook=post_call_hook,
        )

    return override.apply_to_tool(tool)



import pytest

from fastmcp_agents.conversation.types import (
    AssistantConversationEntry,
    CallToolRequest,
    Conversation,
    SystemConversationEntry,
    TextContent,
    ToolConversationEntry,
    UserConversationEntry,
)


@pytest.fixture
def conversation():
    """Create a fresh conversation for each test."""
    return Conversation()


def test_conversation_initialization(conversation: Conversation):
    """Test that a conversation can be initialized."""
    assert conversation is not None
    assert len(conversation.entries) == 0


def test_conversation_add_system_message():
    """Test adding a system message to the conversation."""
    conv: Conversation = Conversation()
    system_entry = SystemConversationEntry(content="You are a test assistant")
    new_conv = conv.append(system_entry)

    assert len(new_conv.entries) == 1
    assert new_conv.entries[0].role == "system"
    assert new_conv.entries[0].content == "You are a test assistant"


def test_conversation_add_user_message():
    """Test adding a user message to the conversation."""
    conv: Conversation = Conversation()
    user_entry = UserConversationEntry(content="Hello")
    new_conv = conv.append(user_entry)

    assert len(new_conv.entries) == 1
    assert new_conv.entries[0].role == "user"
    assert new_conv.entries[0].content == "Hello"


def test_conversation_add_assistant_message_with_tool_calls():
    """Test adding an assistant message with tool calls."""
    conv: Conversation = Conversation()
    tool_calls = [CallToolRequest(id="call_1", name="test_tool", arguments={"param": "value"})]
    assistant_entry = AssistantConversationEntry(content="I'll help you with that", tool_calls=tool_calls)
    new_conv = conv.append(assistant_entry)

    assert len(new_conv.entries) == 1
    assert new_conv.entries[0].role == "assistant"
    assert new_conv.entries[0].content == "I'll help you with that"
    assert len(new_conv.entries[0].tool_calls) == 1
    assert new_conv.entries[0].tool_calls[0].name == "test_tool"


def test_conversation_add_tool_response():
    """Test adding a tool response to the conversation."""
    conv: Conversation = Conversation()
    tool_entry = ToolConversationEntry(tool_call_id="call_1", name="test_tool", content=[TextContent(type="text", text="Tool response")])
    new_conv = conv.append(tool_entry)

    assert len(new_conv.entries) == 1
    assert new_conv.entries[0].role == "tool"
    assert new_conv.entries[0].tool_call_id == "call_1"
    assert new_conv.entries[0].name == "test_tool"
    assert isinstance(new_conv.entries[0].content[0], TextContent)


def test_conversation_merge():
    """Test merging multiple conversations."""
    conv1 = Conversation().append(SystemConversationEntry(content="System message"))
    conv2 = Conversation().append(UserConversationEntry(content="User message"))

    merged = conv1.merge(conv2)
    assert len(merged.entries) == 2
    assert merged.entries[0].role == "system"
    assert merged.entries[1].role == "user"


def test_conversation_to_messages():
    """Test converting conversation to message format."""
    conv: Conversation = Conversation()
    conv = conv.append(SystemConversationEntry(content="System message"))
    conv = conv.append(UserConversationEntry(content="User message"))

    messages = conv.to_messages()
    assert len(messages) == 2
    assert messages[0]["role"] == "system"
    assert messages[1]["role"] == "user"


def test_get_messages(conversation: Conversation):
    """Test retrieving messages from the conversation."""

    # Conversation is immutable, so we need to add to a new conversation
    conversation = conversation.append(UserConversationEntry(content="Hello"))
    conversation = conversation.append(AssistantConversationEntry(content="Hi there!"))

    messages = conversation.get()
    assert len(messages) == 2
    assert messages[0].role == "user"
    assert messages[1].role == "assistant"
    assert messages[0].content == "Hello"
    assert messages[1].content == "Hi there!"



"""Base class for multi-step agents."""

from typing import ParamSpec, TypeVar

from fastmcp import Context
from fastmcp.tools import Tool as FastMCPTool
from pydantic import BaseModel

from fastmcp_agents.agent.single_step import SingleStepAgent
from fastmcp_agents.conversation.memory.base import MemoryFactoryProtocol, MemoryProtocol
from fastmcp_agents.conversation.types import (
    AssistantConversationEntry,
    Conversation,
    UserConversationEntry,
)
from fastmcp_agents.errors.agent import NoResponseError, TaskFailureError

REQUEST_MODEL = TypeVar("REQUEST_MODEL", bound=BaseModel)

SUCCESS_RESPONSE_MODEL = TypeVar("SUCCESS_RESPONSE_MODEL", bound=BaseModel)
ERROR_RESPONSE_MODEL = TypeVar("ERROR_RESPONSE_MODEL", bound=BaseModel)

DEFAULT_STEP_LIMIT = 15
DEFAULT_MAX_PARALLEL_TOOL_CALLS = 5

P = ParamSpec("P")


class MultiStepAgent(SingleStepAgent):
    """A Multi-step agent that can be registered as a tool on a FastMCP server."""

    def __init__(
        self,
        *args,
        memory_factory: MemoryFactoryProtocol,
        max_parallel_tool_calls: int,
        step_limit: int,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.memory_factory = memory_factory
        self.max_parallel_tool_calls = max_parallel_tool_calls
        self.step_limit = step_limit

    async def run_steps(
        self,
        *args,  # noqa: ARG002
        ctx: Context,
        conversation: Conversation,
        tools: list[FastMCPTool],
        step_limit: int,
        success_response_model: type[SUCCESS_RESPONSE_MODEL],
        error_response_model: type[ERROR_RESPONSE_MODEL],
        raise_on_error_response: bool = True,
        **kwargs,  # noqa: ARG002
    ) -> tuple[Conversation, SUCCESS_RESPONSE_MODEL | ERROR_RESPONSE_MODEL]:
        """Run the agent for a given number of steps.

        Args:
            ctx: The context of the agent.
            conversation: The conversation history to send to the LLM.
            tools: The tools to use. If None, the default tools will be used.
            step_limit: The maximum number of steps to perform.
            success_response_model: The model to use for the success response.
            error_response_model: The model to use for the error response.
            raise_on_error_response: Whether to raise an error if the agent fails.

        Returns:
            A tuple of the final conversation and the requested success or error response model.
        """

        # To be set by a callback function.
        result: SUCCESS_RESPONSE_MODEL | ERROR_RESPONSE_MODEL | None = None

        def report_success(**kwargs):
            """Report successful completion of the task."""
            nonlocal result
            result = success_response_model.model_validate(obj=kwargs)

        def report_error(**kwargs):
            """Report failure of the task."""
            nonlocal result
            result = error_response_model.model_validate(obj=kwargs)

        success_tool = FastMCPTool(fn=report_success, name="report_success", parameters=success_response_model.model_json_schema())
        error_tool = FastMCPTool(fn=report_error, name="report_error", parameters=error_response_model.model_json_schema())

        # Add our callback functions to the tools.
        available_tools = [*tools, success_tool, error_tool]

        for i in range(1, step_limit):
            self._logger.info(f"Running step {i} / {step_limit}")

            assistant_conversation_entry, tool_conversation_entries = await self.run_step(
                ctx=ctx,
                prompt=conversation,
                tools=available_tools,
                step_number=i,
                step_limit=step_limit,
            )

            # Add the assistant and tool conversation entries to the conversation.
            conversation = conversation.extend(
                entries=[assistant_conversation_entry, *tool_conversation_entries],
            )

            # If the result is set, return the conversation and the result.
            if result is not None:
                if raise_on_error_response and isinstance(result, error_response_model):
                    raise TaskFailureError(self.name, result)

                return conversation, result

        # Agent failed to record a success or failure response within the step limit.
        raise NoResponseError(agent_name=self.name)

    async def run(
        self,
        ctx: Context,
        instructions: str | Conversation,
        tools: list[FastMCPTool],
        step_limit: int,
        success_response_model: type[SUCCESS_RESPONSE_MODEL],
        error_response_model: type[ERROR_RESPONSE_MODEL],
        raise_on_error_response: bool,
    ) -> tuple[Conversation, SUCCESS_RESPONSE_MODEL | ERROR_RESPONSE_MODEL]:
        """Run the agent.

        Args:
            ctx: The context of the FastMCP Request.
            instructions: The instructions to send to the LLM.
            tools: The tools to use. If None, the default tools will be used.
            step_limit: The maximum number of steps to perform.
            success_response_model: The model to use for the success response.
            error_response_model: The model to use for the error response.
            raise_on_error_response: Whether to raise an error if the agent fails.

        Returns:
            A tuple of the final conversation and the requested success or error response model.
        """

        memory: MemoryProtocol = self.memory_factory()

        conversation = self._prepare_conversation(memory, instructions)

        available_tools: list[FastMCPTool] = tools or self.default_tools

        conversation, completion_result = await self.run_steps(
            ctx=ctx,
            conversation=conversation,
            tools=available_tools,
            step_limit=step_limit or self.step_limit,
            success_response_model=success_response_model,
            error_response_model=error_response_model,
            raise_on_error_response=raise_on_error_response,
        )

        memory.set(conversation)

        return conversation, completion_result

    def _log_total_token_usage(self, conversation: Conversation):
        """Log the total token usage for the conversation."""
        total_tokens = sum(entry.token_usage or 0 for entry in conversation.entries if isinstance(entry, AssistantConversationEntry))
        self._logger.info(f"Total token usage: {total_tokens}")

    def _prepare_conversation(self, memory: MemoryProtocol, instructions: str | Conversation) -> Conversation:
        """Prepare the conversation for the agent. Either by using the conversation history or the instructions."""

        # If there is no conversation history, use the system prompt.
        if not (conversation := memory.get()):
            conversation = self._system_prompt

        # If the instructions are a string, convert them to a conversation entry.
        if isinstance(instructions, str):
            instructions = Conversation(entries=[UserConversationEntry(content=instructions)])

        # Merge the instructions with the conversation history.
        return conversation.merge(instructions)



from pathlib import Path

import pytest

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.types import TextContent
from tests.conftest import evaluate_with_criteria


@pytest.fixture
def server_config_name():
    return "strawgate_filesystem-operations"


class TestFilesystemOperationsAgent:
    @pytest.fixture
    def agent_name(self):
        return "request_filesystem_operations"

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the file was successfully created
        2. that the content was written correctly
        3. that the file exists in the correct location
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_file_creation(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        Create a new file called 'test.txt' in the current directory with the content 'Hello, World!'
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify file was created
        assert (temp_working_dir / "test.txt").exists()
        assert "success" in text_result.lower()
        assert "created" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 1
        assert agent_tool_calls[0].name == "write_file"

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the file was successfully read
        2. that the content was correctly retrieved
        3. that the file metadata was provided
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_file_reading(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        # First create a file to read
        (temp_working_dir / "test.txt").write_text("Hello, World!")

        instructions = """
        Read the contents of the file 'test.txt' and show me its metadata.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify file was read
        assert "hello, world!" in text_result.lower()
        assert "metadata" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 1
        assert agent_tool_calls[0].name == "read_file"

        return agent, instructions, text_result


class TestFilesystemSearchAgent:
    @pytest.fixture
    def agent_name(self):
        return "request_filesystem_search"

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the search was successfully performed
        2. that matching files were found
        3. that file metadata was provided
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_file_search(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        # First create some test files
        (temp_working_dir / "test1.txt").write_text("Hello, World!")
        (temp_working_dir / "test2.txt").write_text("Hello, Python!")
        (temp_working_dir / "other.txt").write_text("Different content")

        instructions = """
        Search for files containing the text 'Hello' in the current directory.
        Show me the file names and their contents.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify search was performed
        assert "test1.txt" in text_result.lower()
        assert "test2.txt" in text_result.lower()
        assert "hello" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 1
        assert agent_tool_calls[0].name == "search_files"

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the directory was successfully listed
        2. that file metadata was provided
        3. that the listing was properly formatted
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_directory_listing(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        # First create some test files
        (temp_working_dir / "test1.txt").write_text("Hello, World!")
        (temp_working_dir / "test2.txt").write_text("Hello, Python!")
        (temp_working_dir / "other.txt").write_text("Different content")

        instructions = """
        List all files in the current directory with their sizes and last modified dates.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify directory was listed
        assert "test1.txt" in text_result.lower()
        assert "test2.txt" in text_result.lower()
        assert "other.txt" in text_result.lower()
        assert "size" in text_result.lower()
        assert "modified" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 1
        assert agent_tool_calls[0].name == "list_directory"

        return agent, instructions, text_result



"""Base classes and protocols for memory."""

from typing import Generic, Protocol, TypeVar

from fastmcp_agents.conversation.types import Conversation, ConversationEntryTypes


class MemoryProtocol(Protocol):
    """A protocol for memory that can be used to store and retrieve conversation history."""

    def add(self, message: ConversationEntryTypes) -> None: ...

    """Add a message to the conversation history."""

    def get(self) -> Conversation: ...

    """Get the current conversation history."""

    def set(self, conversation_history: Conversation) -> None: ...

    """Set the current conversation history."""

    def reset(self) -> None: ...

    """Reset the memory."""

    def persist(self) -> None: ...

    """Persist the memory."""

    def restore(self) -> None: ...

    """Restore the memory."""


MemoryClass = TypeVar("MemoryClass", bound=MemoryProtocol)


class MemoryFactoryProtocol(Protocol):
    """A factory for creating memory instances."""

    def __call__(self) -> MemoryProtocol: ...


class SharedMemoryFactory(MemoryFactoryProtocol, Generic[MemoryClass]):
    """A factory for sharing memory instances."""

    _shared_memory_instance: MemoryClass

    def __init__(self, memory_class: type[MemoryClass]):
        self._memory_class_instance = memory_class()

    def __call__(self) -> MemoryClass:
        """Return the shared memory instance."""
        return self._memory_class_instance


class PrivateMemoryFactory(MemoryFactoryProtocol, Generic[MemoryClass]):
    """Every call to the PrivateMemoryFactory will return a new memory instance."""

    _memory_class: type[MemoryClass]

    def __init__(self, memory_class: type[MemoryClass]):
        self._memory_class = memory_class

    def __call__(self) -> MemoryClass:
        """Return a new memory instance."""
        return self._memory_class()



"""LiteLLM integration for FastMCP Agents."""

import json
import os

from fastmcp.tools import Tool as FastMCPTool
from litellm import CustomStreamWrapper, LiteLLM, acompletion
from litellm.types.utils import ChatCompletionMessageToolCall, Function, Message, ModelResponse, StreamingChoices
from litellm.utils import supports_function_calling

from fastmcp_agents.conversation.types import AssistantConversationEntry, CallToolRequest, Conversation
from fastmcp_agents.errors.base import NoResponseError, UnknownToolCallError, UnsupportedFeatureError
from fastmcp_agents.errors.llm_link import ModelDoesNotSupportFunctionCallingError, ModelNotSetError
from fastmcp_agents.llm_link.base import (
    AsyncLLMLink,
    CompletionMetadata,
)
from fastmcp_agents.llm_link.utils import transform_fastmcp_tool_to_openai_tool


class AsyncLitellmLLMLink(AsyncLLMLink):
    model: str

    def __init__(self, model: str | None = None, completion_kwargs: dict | None = None, client: LiteLLM | None = None) -> None:
        """Create a new Litellm LLM link.

        Args:
            model: The model to use.
            completion_kwargs: The completion kwargs to use.
            client: The Litellm client to use.
        """
        self.client = client or LiteLLM()

        if model := (model or os.getenv("MODEL")):
            self.model = model
        else:
            raise ModelNotSetError

        self.completion_kwargs = completion_kwargs or {}

        self.validate_model(model)

    @classmethod
    def validate_model(cls, model: str):
        """Validate that the model is a valid Litellm model and that it supports function calling

        Args:
            model: The model to validate.
        """
        if not supports_function_calling(model=model):
            raise ModelDoesNotSupportFunctionCallingError(model=model)

    def _extract_tool_calls(self, message: Message) -> list[CallToolRequest]:
        """Extract the tool calls from the message.

        Args:
            message: The message to extract the tool calls from.

        Returns:
            A list of tool calls.
        """
        if not (tool_calls := message.tool_calls):
            raise NoResponseError(missing_item="tool calls", model=self.model)

        self.logger.debug(f"Response contains {len(tool_calls)} tool requests: {tool_calls}")

        tool_call_requests: list[CallToolRequest] = []

        for tool_call in tool_calls:
            if not isinstance(tool_call, ChatCompletionMessageToolCall):
                raise UnknownToolCallError(
                    tool_name=tool_call.name, extra_info=f"Tool call is not a ChatCompletionMessageToolCall: {tool_call}"
                )

            tool_call_function: Function = tool_call.function

            if not tool_call_function.name:
                raise UnknownToolCallError(tool_name="unknown", extra_info=f"Tool call has no name: {tool_call}")

            cast_arguments = json.loads(tool_call_function.arguments) or {}

            tool_call_requests.append(
                CallToolRequest(
                    id=tool_call.id,
                    name=tool_call_function.name,
                    arguments=cast_arguments,
                )
            )

        return tool_call_requests

    def _extract_message(self, response: ModelResponse) -> Message:
        """Extract the response message from the response. This contains the tool_calls and content."""
        if not (choices := response.choices) or len(choices) == 0:
            raise NoResponseError(missing_item="choices", model=self.model)

        if not (choice := choices[0]):
            raise NoResponseError(missing_item="choice", model=self.model)

        if isinstance(choice, StreamingChoices):
            raise UnsupportedFeatureError(feature="streaming completions")

        if not (chosen_message := choice.message):
            raise NoResponseError(missing_item="response message", model=self.model)

        return chosen_message

    async def _extract_tool_call_requests(self, response: ModelResponse) -> list[CallToolRequest]:
        """Extract the tool calls from the response.

        Returns:
            A list of tool call requests.
        """

        message = self._extract_message(response)

        return self._extract_tool_calls(message)

    async def async_completion(
        self,
        conversation: Conversation,
        fastmcp_tools: list[FastMCPTool],
    ) -> AssistantConversationEntry:
        """Call the LLM with the given messages and tools.

        Args:
            messages: The messages to send to the LLM.
            tools: The tools to use.

        Returns:
            The assistant conversation entry.
        """

        litellm_tools = [transform_fastmcp_tool_to_openai_tool(tool) for tool in fastmcp_tools]

        messages = conversation.to_messages()

        model_response = await acompletion(
            messages=messages,
            model=self.model,
            **self.completion_kwargs,
            tools=litellm_tools,
            tool_choice="required",
            num_retries=3,
        )

        # Make the typechecker happy
        if isinstance(model_response, CustomStreamWrapper):
            raise UnsupportedFeatureError(feature="streaming completions")

        completion_metadata = CompletionMetadata()
        if model_response.model_extra:
            token_usage = model_response.model_extra.get("usage", {}).get("total_tokens", 0)

            completion_metadata.token_usage = token_usage
            self.token_usage += token_usage

        message = self._extract_message(model_response)

        tool_call_requests = self._extract_tool_calls(message)

        return AssistantConversationEntry(role="assistant", tool_calls=tool_call_requests, token_usage=completion_metadata.token_usage)



"""Custom errors for the loader."""

from fastmcp_agents.errors.base import FastMCPAgentsError


class LoaderError(FastMCPAgentsError):
    """Raised when there is an error loading the agent."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ToolsNotFoundError(LoaderError):
    """Raised when a tool is not found."""

    def __init__(self, agent_name: str, missing_tools: set[str]):
        self.message = f"Agent '{agent_name}' is missing {len(missing_tools)} tools: {missing_tools}"
        super().__init__(self.message)


class MCPServerStartupError(LoaderError):
    """Raised when a MCP server fails to start."""

    def __init__(self, server_name: str, error: Exception):
        self.message = f"MCP server '{server_name}' failed to start: {error}"
        super().__init__(self.message)



"""Custom errors for the CLI."""

from fastmcp_agents.errors.base import FastMCPAgentsError


class NoServerToWrapError(FastMCPAgentsError):
    """Raised when no server is provided to wrap."""

    def __init__(self):
        self.message = "No server provided to wrap"
        super().__init__(self.message)


class MultipleConfigOptionsError(FastMCPAgentsError):
    """Raised when multiple config options are provided."""

    def __init__(self):
        self.message = "Only one of config-file, config-url, or config-bundled can be provided"
        super().__init__(self.message)


class NoConfigError(FastMCPAgentsError):
    """Raised when no config is provided."""

    def __init__(self):
        self.message = "No config provided"
        super().__init__(self.message)


class MCPServerError(FastMCPAgentsError):
    """Raised when an MCP server fails to wrap."""

    def __init__(self, server: str, error: Exception):
        self.message = f"Error wrapping MCP server {server}: {error}"
        super().__init__(self.message)



"""Logging configuration for FastMCP Agents."""

import logging
from typing import Literal

from fastmcp.utilities.logging import get_logger
from rich.console import Console
from rich.logging import RichHandler

BASE_LOGGER = get_logger("agents")


def setup_logging(
    level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] | int = "INFO",
):
    handler = RichHandler(
        console=Console(stderr=True),
        rich_tracebacks=False,
    )
    formatter = logging.Formatter("%(name)s : %(message)s")
    handler.setFormatter(formatter)

    BASE_LOGGER.setLevel(level)
    BASE_LOGGER.addHandler(handler)
    BASE_LOGGER.propagate = False

    logging.getLogger("mcp").setLevel("WARNING")


def get_logger(name: str) -> logging.Logger:
    """Get a logger nested under FastMCP namespace.

    Args:
        name: the name of the logger, which will be prefixed with 'FastMCP.'

    Returns:
        a configured logger instance
    """
    return BASE_LOGGER.getChild(name)



"""Custom errors for agents."""

from litellm import BaseModel

from fastmcp_agents.errors.base import FastMCPAgentsError


class ToolNotFoundError(FastMCPAgentsError):
    def __init__(self, agent_name: str, tool_name: str):
        self.message = f"Agent '{agent_name}' tried calling tool '{tool_name}' but it was not found"
        super().__init__(self.message)


class NoResponseError(FastMCPAgentsError):
    def __init__(self, agent_name: str):
        self.message = f"Agent '{agent_name}' did not return a response"
        super().__init__(self.message)


class TaskFailureError(FastMCPAgentsError):
    def __init__(self, agent_name: str, error: BaseModel):
        self.message = f"Agent '{agent_name}' failed to complete the task: {error}"
        super().__init__(self.message)



"""Custom errors for tool transformation."""


class TransformedToolError(Exception):
    """An error that occurs when a tool is transformed."""

    def __init__(self, message: str):
        super().__init__(message)


class ToolParameterOverrideError(TransformedToolError):
    """An error that occurs when a parameter override is invalid."""

    def __init__(self, parameter_name: str):
        super().__init__(f"Parameter {parameter_name} not found in tool.")


class ToolParameterNotFoundError(TransformedToolError):
    """An error that occurs when a parameter is not found in the schema."""

    def __init__(self, parameter_name: str):
        super().__init__(f"Parameter {parameter_name} not found in tool.")


class ToolParameterAlreadyExistsError(TransformedToolError):
    """An error that occurs when a parameter already exists in the schema."""

    def __init__(self, parameter_name: str):
        super().__init__(f"Parameter {parameter_name} already exists in schema.")



"""Base classes and protocols for LLM links."""

import logging
from typing import Any, Protocol

from fastmcp.tools import Tool as FastMCPTool
from pydantic import BaseModel, Field

from fastmcp_agents.conversation.types import AssistantConversationEntry, Conversation
from fastmcp_agents.observability.logging import BASE_LOGGER

logger = BASE_LOGGER.getChild("llm_link")


class CompletionMetadata(BaseModel):
    """Metadata about the completion."""

    token_usage: int | None = Field(default=None, description="The number of tokens used by the LLM.")
    """The number of tokens used by the LLM."""


class AsyncLLMLink(Protocol):
    """Base class for all LLM links.

    This class is used to abstract the LLM link implementation from the agent.
    """

    completion_kwargs: dict[str, Any]
    """The kwargs to pass to the underlying LLM SDK when asking for a completion."""

    token_usage: int = 0
    """The number of tokens used by the LLM."""

    logger: logging.Logger = logger
    """The logger to use for the LLM link."""

    async def async_completion(
        self,
        conversation: Conversation,
        fastmcp_tools: list[FastMCPTool],
    ) -> AssistantConversationEntry: ...

    """Call the LLM with the given messages and tools.

    Args:
        conversation: The conversation to send to the LLM.
        fastmcp_tools: The tools to use.

    Returns:
        The assistant conversation entry.
    """



from pathlib import Path

import pytest

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.conversation.types import TextContent
from tests.conftest import evaluate_with_criteria


@pytest.fixture
def server_config_name():
    return "github_github-mcp-server"


class TestGitHubAgent:
    @pytest.fixture
    def agent_name(self):
        return "summarize_github_issue"

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the issue was successfully retrieved
        2. that the issue comments were retrieved
        3. that a clear summary of the issue and comments was provided
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_issue_summarization(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        Summarize issue #1 in the repository modelcontextprotocol/servers.
        Include any relevant comments and provide a clear overview of the issue's status and content.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify issue was retrieved and summarized
        assert "issue" in text_result.lower()
        assert "summary" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 2
        assert agent_tool_calls[0].name == "get_issue"
        assert agent_tool_calls[1].name == "get_issue_comments"

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the issue was successfully retrieved
        2. that related issues were searched for
        3. that a clear summary of related issues was provided
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_related_issues(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        Find issues related to issue #1 in the repository modelcontextprotocol/servers.
        Include a confidence rating for each related issue and explain why it's related.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify related issues were found
        assert "related" in text_result.lower()
        assert "confidence" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 2
        assert agent_tool_calls[0].name == "get_issue"
        assert agent_tool_calls[1].name == "search_issues"

        return agent, instructions, text_result


class TestPullRequestAgent:
    @pytest.fixture
    def agent_name(self):
        return "summarize_pull_request"

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the pull request was successfully retrieved
        2. that the PR files were retrieved
        3. that the PR status was checked
        4. that a clear summary of the PR was provided
        5. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_pr_summarization(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        Summarize pull request #1 in the repository modelcontextprotocol/servers.
        Include the files changed, review status, and any relevant comments.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify PR was retrieved and summarized
        assert "pull request" in text_result.lower()
        assert "summary" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 4
        assert agent_tool_calls[0].name == "get_pull_request"
        assert agent_tool_calls[1].name == "get_pull_request_files"
        assert agent_tool_calls[2].name == "get_pull_request_status"
        assert agent_tool_calls[3].name == "get_pull_request_comments"

        return agent, instructions, text_result

    @evaluate_with_criteria(
        criteria="""
        The result and the conversation history must indicate:
        1. that the pull request was successfully retrieved
        2. that the PR reviews were retrieved
        3. that a clear summary of the review status was provided
        4. that the agent used the correct sequence of tools

        Any other response is a failure.
        """,
        minimum_grade=0.9,
    )
    async def test_pr_reviews(self, temp_working_dir: Path, agent: FastMCPAgent, call_curator, agent_tool_calls):
        instructions = """
        Summarize the review status of pull request #1 in the repository modelcontextprotocol/servers.
        Include the review status, reviewer comments, and any requested changes.
        """

        result = await call_curator(name=agent.name, instructions=instructions)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TextContent)
        text_result = result[0].text

        # Verify reviews were retrieved and summarized
        assert "review" in text_result.lower()
        assert "status" in text_result.lower()

        # Verify tool calls
        assert len(agent_tool_calls) >= 2
        assert agent_tool_calls[0].name == "get_pull_request"
        assert agent_tool_calls[1].name == "get_pull_request_reviews"

        return agent, instructions, text_result



"""Sample code for FastMCP using InterceptingProxyTool."""

import asyncio

from fastmcp import FastMCP
from fastmcp.client import Client
from fastmcp_agents.vendored.tool_transformer.models import ToolParameter
from fastmcp_agents.vendored.tool_transformer.tool_transformer import proxy_tool

third_party_mcp_config = {
    "time": {
        "command": "uvx",
        "args": [
            "git+https://github.com/modelcontextprotocol/servers.git@2025.4.24#subdirectory=src/time",
            "--local-timezone=America/New_York",
        ],
    }
}


async def async_main():
    async with Client(third_party_mcp_config, timeout=30) as remote_mcp_client:
        proxied_mcp_server = FastMCP.as_proxy(remote_mcp_client)

        proxied_tools = await proxied_mcp_server.get_tools()

        frontend_server = FastMCP("Frontend Server")

        proxy_tool(
            proxied_tools["convert_time"],
            server=frontend_server,
            name="transformed_convert_time",
            description="Converts a time from New York to another timezone.",
            parameter_overrides=[
                ToolParameter[str](
                    name="source_timezone",
                    description="The timezone of the time to convert.",
                    constant="America/New_York",  # Source Timezone is now required to be America/New_York
                ),
                ToolParameter[str](
                    name="time",
                    description="The time to convert. Must be in the format HH:MM. Default is 3:00.",
                    default="3:00",  # Time now defaults to 3:00
                ),
                # No override of the override the target_timezone parameter
            ],
        )

        await frontend_server.run_async(transport="sse")


def run_mcp():
    asyncio.run(async_main())


if __name__ == "__main__":
    run_mcp()



from fastmcp_agents.vendored.tool_transformer.models import (
    BooleanToolParameter,
    FloatToolParameter,
    IntToolParameter,
    PostToolCallHookProtocol,
    PreToolCallHookProtocol,
    StringToolParameter,
    ToolOverride,
    ToolParameter,
    ToolParameterTypes,
)
from fastmcp_agents.vendored.tool_transformer.tool_transformer import transform_tool

__all__ = [
    "BooleanToolParameter",
    "FloatToolParameter",
    "IntToolParameter",
    "PostToolCallHookProtocol",
    "PreToolCallHookProtocol",
    "StringToolParameter",
    "ToolOverride",
    "ToolParameter",
    "ToolParameterTypes",
    "transform_tool",
]



"""FastMCP Agent implementation."""

from typing import TypeAlias

from fastmcp import Context
from fastmcp.tools import Tool as FastMCPTool
from pydantic import BaseModel, Field

from fastmcp_agents.agent.multi_step import ERROR_RESPONSE_MODEL, SUCCESS_RESPONSE_MODEL, MultiStepAgent
from fastmcp_agents.conversation.memory.base import MemoryFactoryProtocol, PrivateMemoryFactory
from fastmcp_agents.conversation.memory.ephemeral import EphemeralMemory
from fastmcp_agents.conversation.types import (
    Conversation,
    SystemConversationEntry,
)
from fastmcp_agents.errors.agent import TaskFailureError
from fastmcp_agents.llm_link.base import AsyncLLMLink
from fastmcp_agents.llm_link.litellm import AsyncLitellmLLMLink

DEFAULT_SYSTEM_PROMPT = """
You are `{agent_name}`, an AI Agent that is embedded into a FastMCP Server. You act as an
interface between the remote user/agent and the tools available on the server.

The person or Agent that invoked you understood received the following as a description of your capabilities:

````````markdown
{agent_description}
````````

You will be asked to perform a task. Your tasks will be phrased in the form of either `tell {agent_name} to <task>` or just `<task>`.
You should leverage the tools available to you to perform the task. You may perform many tool calls in one go but you should always
keep in mind that the tool calls may run in any order and may run at the same time. So you should plan
for which calls you can do in parallel (multiple in a single request) and which you should do sequentially (one tool call per request).
"""

DEFAULT_STEP_LIMIT = 20
"""The default step limit for the agent."""

DEFAULT_MAX_PARALLEL_TOOL_CALLS = 5
"""The default maximum number of tool calls to perform in parallel."""


class DefaultErrorResponseModel(BaseModel):
    """A default error response model for the agent."""

    error: str = Field(..., description="The error message if the agent failed. You must provide a string error message.")


class DefaultSuccessResponseModel(BaseModel):
    """A default success response model for the agent."""

    success: bool = Field(..., description="Whether the agent was successful")
    result: str = Field(..., description="The result of the agent. You must provide a string result.")


DefaultResponseModelTypes: TypeAlias = DefaultErrorResponseModel | DefaultSuccessResponseModel


class FastMCPAgent(MultiStepAgent):
    """A basic FastMCP Agent that can be used to perform tasks.

    This agent is a multi-step agent that can be used to perform tasks. It will use the provided LLM link to perform the tasks.
    It will also use the provided memory factory to store the conversation history.
    """

    def __init__(
        self,
        *,
        name: str,
        description: str,
        system_prompt: str | Conversation | None = None,
        default_tools: list[FastMCPTool] | None = None,
        llm_link: AsyncLLMLink | None = None,
        memory_factory: MemoryFactoryProtocol | None = None,
        max_parallel_tool_calls: int | None = None,
        step_limit: int | None = None,
        **kwargs,
    ):
        """Create a new FastMCP Agent.

        Args:
            name: The name of the agent.
            description: The description of the agent.
            instructions: The instructions for the agent.
            default_tools: The default tools to use. Defaults to an empty list.
            llm_link: The LLM link to use. Defaults to a Litellm LLM link.
            system_prompt: The system prompt to use. Defaults to a default system prompt.
            memory_factory: The memory factory to use. Defaults to a private memory factory.
            max_parallel_tool_calls: The maximum number of tool calls to perform in parallel. Defaults to 5.
            step_limit: The maximum number of steps to perform. Defaults to 10.
        """

        if system_prompt is None:
            system_prompt = DEFAULT_SYSTEM_PROMPT

        if isinstance(system_prompt, str):
            system_prompt = system_prompt.format(agent_name=name, agent_description=description)

            system_prompt = Conversation(entries=[SystemConversationEntry(content=system_prompt)])

        super().__init__(
            name=name,
            description=description,
            system_prompt=system_prompt,
            llm_link=llm_link or AsyncLitellmLLMLink(),
            default_tools=default_tools or [],
            memory_factory=memory_factory or PrivateMemoryFactory(memory_class=EphemeralMemory),
            max_parallel_tool_calls=max_parallel_tool_calls or DEFAULT_MAX_PARALLEL_TOOL_CALLS,
            step_limit=step_limit or DEFAULT_STEP_LIMIT,
            **kwargs,
        )

    async def run(
        self,
        ctx: Context,
        instructions: str | Conversation,
        tools: list[FastMCPTool] | None = None,
        step_limit: int | None = None,
        success_response_model: type[SUCCESS_RESPONSE_MODEL] = DefaultSuccessResponseModel,
        error_response_model: type[ERROR_RESPONSE_MODEL] = DefaultErrorResponseModel,
        raise_on_error_response: bool = True,
    ) -> tuple[Conversation, SUCCESS_RESPONSE_MODEL | ERROR_RESPONSE_MODEL]:
        return await super().run(
            ctx=ctx,
            instructions=instructions,
            tools=tools or [],
            step_limit=step_limit or DEFAULT_STEP_LIMIT,
            success_response_model=success_response_model,
            error_response_model=error_response_model,
            raise_on_error_response=raise_on_error_response,
        )

    async def currate(self, ctx: Context, instructions: str) -> str:
        """Runs the agent with the provided instructions and default tools, returning a string result or raising a TaskFailureError.

        Useful for making the Agent available as a general purpose tool on the server.s
        """

        _, result = await self.run(ctx, instructions, self.default_tools, self.step_limit)

        if isinstance(result, DefaultErrorResponseModel):
            raise TaskFailureError(self.name, result)

        return result.result



"""Disk-backed memory implementation."""

from pathlib import Path

import yaml
from pydantic import Field

from fastmcp_agents.conversation.memory.base import Conversation, MemoryProtocol
from fastmcp_agents.conversation.types import ConversationEntryTypes
from fastmcp_agents.errors.base import ContributionsWelcomeError


class DiskBackedMemory(MemoryProtocol):
    """
    A memory entry that is backed by a disk file.

    NOTE: This feature is currently unimplemented. Contributions are welcome!
    """

    conversation_history: Conversation = Field(default_factory=Conversation)
    path: Path

    def __init__(self, path: Path | str):
        """
        Initializes the DiskBackedMemory.

        Args:
            path: The path to the file to store the conversation history.

        Raises:
            ContributionsWelcomeError: This feature is currently unimplemented.
        """
        raise ContributionsWelcomeError(feature="DiskBackedMemory")
        self.path = Path(path)

        if self.path.exists():
            self.restore()
        else:
            self.conversation_history = Conversation()

    def add(self, message: ConversationEntryTypes):
        self.conversation_history.append(message)

    def get(self):
        return self.conversation_history.get()

    def reset(self) -> None:
        pass

    def persist(self) -> None:
        with self.path.open("w") as f:
            conversation_history_dict = [conversation_entry.model_dump() for conversation_entry in self.conversation_history.get()]

            yaml.safe_dump(conversation_history_dict, f, width=140)

    def restore(self):
        with self.path.open("r") as f:
            conversation_history_dict = yaml.safe_load(f)

            self.conversation_history = Conversation.model_validate(**conversation_history_dict)



"""Base CLI for FastMCP Agents."""

import asyncio
import json
import os
import sys
from typing import Any, Literal

import asyncclick as click
from fastmcp import Client, FastMCP
from fastmcp.exceptions import ToolError
from fastmcp.utilities.logging import configure_logging
from mcp.types import EmbeddedResource, ImageContent, TextContent
from pydantic import BaseModel, Field

from fastmcp_agents.agent.fastmcp import FastMCPAgent
from fastmcp_agents.cli.loader import get_config_for_bundled, get_config_from_file, get_config_from_url
from fastmcp_agents.cli.models import (
    AgentModel,
    AugmentedServerModel,
    OverriddenStdioMCPServer,
    ServerSettings,
)
from fastmcp_agents.errors.base import ContributionsWelcomeError
from fastmcp_agents.errors.cli import FastMCPAgentsError, NoConfigError
from fastmcp_agents.observability.logging import get_logger, setup_logging

logger = get_logger("cli")

if os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT") is not None:
    from fastmcp_agents.observability.otel import setup_otel

    setup_otel()


class PendingToolCall(BaseModel):
    """A pending tool call. Only used in the CLI class."""

    name: str
    arguments: dict[str, Any]


class ToolCallResult(PendingToolCall):
    """A result of a tool call. Only used in the CLI class."""

    name: str
    arguments: dict[str, Any]
    result: list[TextContent | ImageContent | EmbeddedResource]


class CliContext(BaseModel):
    """A context object that gets passed around the CLI commands."""

    server_settings: ServerSettings
    augmented_server_model: AugmentedServerModel = Field(default_factory=AugmentedServerModel)
    pending_tool_calls: list[PendingToolCall] = Field(default_factory=list)


@click.group()
@click.option(
    "--transport",
    type=click.Choice(["stdio", "sse", "streamable-http"]),
    default="stdio",
    help="The transport to use for the MCP server. (stdio, sse, streamable-http)",
)
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO",
    help="The logging level to use for the agent.",
)
@click.option("--agent-only", is_flag=True, help="Only run the agents, don't expose the tools to the client.")
@click.option("--tool-only", is_flag=True, help="Only run the tools, don't expose the agents to the client.")
@click.pass_context
def cli_base(
    ctx: click.Context,
    transport: Literal["stdio", "sse", "streamable-http"],
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    agent_only: bool = False,
    tool_only: bool = False,
):
    """The base CLI for the FastMCP Agents. Configure logging, transport, and filtering."""

    # Setup FastMCP Agents Logging
    setup_logging(level=log_level)

    # Setup FastMCP Logging
    configure_logging(level=log_level, enable_rich_tracebacks=False)

    ctx.obj = CliContext(
        server_settings=ServerSettings(
            transport=transport,
            log_level=log_level,
            agent_only=agent_only,
            tool_only=tool_only,
        )
    )


@cli_base.group(name="cli", chain=True)
def cli_interface():
    """The group for the `CLI` part of the FastMCP Agents CLI.

    This allows building agents and wrapping servers entirely from the command line.
    """


@cli_interface.command(name="agent")
@click.option("--name", help="The name of the agent to wrap.")
@click.option("--description", help="The description of the agent to wrap.")
@click.option("--instructions", help="The instructions of the agent to wrap.")
@click.option("--allowed-tools", help="A comma separated list of the tools that the agent can use.")
@click.option("--blocked-tools", help="A comma separated list of the tools that the agent cannot use.")
@click.pass_context
def cli_build_agent(
    ctx: click.Context,
    name: str,
    description: str,
    instructions: str,
    allowed_tools: str,
    blocked_tools: str,
):
    """Build an agent over the command line."""
    split_allowed_tools = allowed_tools.split(",") if allowed_tools else None
    split_blocked_tools = blocked_tools.split(",") if blocked_tools else None

    cli_context: CliContext = ctx.obj

    cli_context.augmented_server_model.agents.append(
        AgentModel(
            name=name,
            description=description,
            default_instructions=instructions,
            allowed_tools=split_allowed_tools,
            blocked_tools=split_blocked_tools,
        )
    )


@cli_base.group(name="config", chain=True)
@click.option("--file", type=str, help="The config file to use.")
@click.option("--url", type=str, help="The URL of the config file to use.")
@click.option(
    "--directory",
    envvar="FASTMCP_AGENTS_CONFIG_DIR",
    type=click.Path(exists=True),
    help="A directory of config files, from which --file is relative to",
)
@click.option("--bundled", type=str, help="The bundled server to use.")
@click.pass_context
def cli_with_config(
    ctx: click.Context, file: str | None = None, url: str | None = None, bundled: str | None = None, directory: str | None = None
):
    """Load a config file from the command line."""

    cli_context: CliContext = ctx.obj

    if url:
        cli_context.augmented_server_model = get_config_from_url(url)
    elif file:
        cli_context.augmented_server_model = get_config_from_file(directory=directory, file=file)
    elif bundled:
        cli_context.augmented_server_model = get_config_for_bundled(bundled)
    else:
        raise NoConfigError


async def handle_pending_tool_calls(
    mcp_clients: list[Client], server: FastMCP, pending_tool_calls: list[PendingToolCall]
) -> list[ToolCallResult]:
    results: list[ToolCallResult] = []

    server_client = Client(transport=server)

    async with server_client:
        for pending_tool_call in pending_tool_calls:
            try:
                result = await server_client.call_tool(pending_tool_call.name, pending_tool_call.arguments)
                tool_call_result = ToolCallResult(name=pending_tool_call.name, arguments=pending_tool_call.arguments, result=result)
                results.append(tool_call_result)
            except ToolError:  # noqa: PERF203
                logger.error(f"Tool {pending_tool_call.name} with arguments {pending_tool_call.arguments} returned an error.")  # noqa: TRY400

    for mcp_client in mcp_clients:
        await mcp_client.close()

    for result in results:
        logger.info(f"Tool {result.name} with arguments {result.arguments} returned result:\n{result.result}")

    return results


async def run_server_or_call_tools(
    agents: list[FastMCPAgent],
    mcp_clients: list[Client],
    server: FastMCP,
    pending_tool_calls: list[PendingToolCall],
    transport: Literal["stdio", "sse", "streamable-http"],
):
    """A shared helper for either running the server or handling pending tool calls."""
    if pending_tool_calls:
        await handle_pending_tool_calls(mcp_clients, server, pending_tool_calls)
    else:
        await server.run_async(transport=transport)

    total_token_usage = sum(agent.llm_link.token_usage for agent in agents)
    logger.info(f"Total token usage: {total_token_usage}")


@cli_with_config.command(name="run")
@click.pass_context
async def cli_with_config_run(ctx: click.Context):
    """Run the server with configuration from a config source."""
    cli_context: CliContext = ctx.obj

    agents, mcp_clients, server = await cli_context.augmented_server_model.to_fastmcp_server(server_settings=cli_context.server_settings)

    await run_server_or_call_tools(
        agents=agents,
        mcp_clients=mcp_clients,
        server=server,
        pending_tool_calls=cli_context.pending_tool_calls,
        transport=cli_context.server_settings.transport,
    )


@cli_interface.command(name="list")
@click.pass_context
async def list_tools(
    ctx: click.Context,
):
    """List the tools available on the server."""
    cli_context: CliContext = ctx.obj

    _, mcp_clients, server = await cli_context.augmented_server_model.to_fastmcp_server(server_settings=cli_context.server_settings)

    tools = await server.get_tools()

    for mcp_client in mcp_clients:
        await mcp_client.close()

    logger.info("Listing tools:")
    for tool in tools.values():
        logger.info(f"Tool: {tool.name}")


@cli_interface.command(name="call", context_settings={"ignore_unknown_options": True, "allow_extra_args": True})
@click.argument("name", type=str)
@click.argument("parameters", type=str)
@click.pass_context
def call_tool(
    ctx: click.Context,
    name: str,
    parameters: str,
):
    """Add a tool call to the pending tool calls list."""
    cli_context: CliContext = ctx.obj

    arguments = json.loads(parameters)

    cli_context.pending_tool_calls.append(PendingToolCall(name=name, arguments=arguments))


@cli_base.command(name="shell")
@click.pass_context
def shell(ctx: click.Context):  # noqa: ARG001
    """
    Start a shell session with the server.

    NOTE: This feature is currently unimplemented. Contributions are welcome!
    """
    raise ContributionsWelcomeError(feature="shell")


@cli_interface.command(name="wrap", context_settings={"ignore_unknown_options": True, "allow_extra_args": True})
@click.option("--env", type=str, multiple=True, help="The environment variables to set for the server.")
@click.argument("direct-wrap-args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
async def wrap_server_run_mcp(
    ctx: click.Context,
    env: list[str],
    direct_wrap_args: list[str],
):
    """Take the last of the cli args and use them to run the mcp server and run it."""
    cli_context: CliContext = ctx.obj

    command = direct_wrap_args[0]
    args = direct_wrap_args[1:] if len(direct_wrap_args) > 1 else []
    environment = os.environ.copy()

    for env_var in env:
        key, value = env_var.split("=")
        environment[key] = value

    cli_context.augmented_server_model.mcpServers = {
        "main": OverriddenStdioMCPServer(
            command=command,
            args=args,
            env=environment,
        )
    }

    agents, mcp_clients, server = await cli_context.augmented_server_model.to_fastmcp_server(server_settings=cli_context.server_settings)

    await run_server_or_call_tools(
        agents=agents,
        mcp_clients=mcp_clients,
        server=server,
        pending_tool_calls=cli_context.pending_tool_calls,
        transport=cli_context.server_settings.transport,
    )


cli_with_config.add_command(call_tool)
cli_with_config.add_command(list_tools)


def run_mcp():
    try:
        asyncio.run(cli_base())
    except FastMCPAgentsError:
        logger.exception(msg="An error occurred while running FastMCPAgents")
        sys.exit(1)
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt received, shutting down...")
        sys.exit(0)
    except Exception:
        logger.exception(msg="An unknown error occurred while running FastMCPAgents")
        sys.exit(1)


if __name__ == "__main__":
    run_mcp()
